# -*- coding: utf-8 -*-
"""
Functions related to fetching and manipulating skeletons.
"""
from io import StringIO
from itertools import combinations
from collections import namedtuple

import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial import cKDTree

from .client import inject_client


@inject_client
def fetch_skeleton(body, heal=False, export_path=None, format='pandas', with_distances=False, *, client=None):
    """
    Equivalent to :py:meth:`.Client.fetch_skeleton()`.  See that function for details.
    """
    return client.fetch_skeleton(body, heal, export_path, format, with_distances)


def skeleton_df_to_nx(df, with_attributes=True, directed=True, with_distances=False):
    """
    Create a ``networkx.Graph`` from a skeleton DataFrame.

    Args:
        df:
            DataFrame as returned by :py:meth:`.Client.fetch_skeleton()`

        with_attributes:
            If True, store node attributes for x, y, z, radius

        directed:
            If True, return ``nx.DiGraph``, otherwise ``nx.Graph``.
            Edges will point from child to parent.

        with_distances:
            If True, add an edge attribute 'distance' indicating the
            euclidean distance between skeleton nodes.

    Returns:
        ``nx.DiGraph`` or ``nx.Graph``
    """
    if directed:
        g = nx.DiGraph()
    else:
        g = nx.Graph()

    if with_attributes:
        for row in df.itertuples(index=False):
            g.add_node(row.rowId, x=row.x, y=row.y, z=row.z, radius=row.radius)
    else:
        g.add_nodes_from(df['rowId'].sort_values())

    # Instead of assuming that the root node refers to a special parent (e.g. -1),
    # we determine the root_parents by inspection.
    root_parents = pd.Index(df['link'].unique()).difference(df['rowId'].unique())
    root_parents

    if with_distances:
        edges_df = df[['rowId', 'link']].copy()
        edges_df['distance'] = calc_segment_distances(df)
        edges_df = edges_df.query('link not in @root_parents').sort_values(['rowId', 'link'])
        g.add_weighted_edges_from(edges_df.itertuples(index=False), 'distance')
    else:
        edges_df = df.query('link not in @root_parents')[['rowId', 'link']]
        edges_df = edges_df.sort_values(['rowId', 'link'])
        g.add_edges_from(edges_df.values)

    return g


def calc_segment_distances(df):
    """
    For each node (row) in the given skeleton DataFrame,
    compute euclidean distance from the node to its parent (link) node.
    Root nodes (i.e. when link == -1) will be assigned a distance of np.inf.

    Returns:
        np.ndarray
    """
    # Append parent (link) columns to each row by matching
    # each row's 'link' ID with the parent's 'rowId'.
    edges_df = df[['rowId', 'link', *'xyz']].merge(
        df[['rowId', *'xyz']], 'left',
        left_on='link', right_on='rowId', suffixes=['', '_link'])

    diff = edges_df[[*'xyz']] - edges_df[['x_link', 'y_link', 'z_link']].values
    distances = np.linalg.norm(diff, axis=1).astype(np.float32)
    distances[np.isnan(distances)] = np.inf
    return distances


def skeleton_swc_to_df(swc):
    """
    Create a DataFrame from and SWC file.
    The 'node_type' column is discarded.

    Args:
        swc:
            Either a filepath ending in '.swc', or a file object,
            or the contents of an SWC file (as a string).

    Returns:
        ``pd.DataFrame``
    """
    if hasattr(swc, 'read'):
        swc = swc.read()
    else:
        assert isinstance(swc, str)
        if swc.endswith('.swc'):
            with open(swc, 'r') as f:
                swc = f.read()

    cols = ['rowId', 'node_type', 'x', 'y', 'z', 'radius', 'link']
    lines = swc.split('\n')
    lines = filter(lambda line: '#' not in line, lines)
    swc_csv = '\n'.join(lines)

    # Compact dtypes save RAM when loading lots of skeletons
    dtypes = {
        'rowId': np.int32,
        'node_type': np.int8,
        'x': np.float32,
        'y': np.float32,
        'z': np.float32,
        'radius': np.float32,
        'link': np.int32,
    }
    df = pd.read_csv(StringIO(swc_csv), delimiter=' ', engine='c', names=cols, dtype=dtypes, header=None)
    df = df.drop(columns=['node_type'])
    return df


def skeleton_df_to_swc(df, export_path=None):
    """
    Create an SWC file from a skeleton DataFrame.

    Args:
        df:
            DataFrame, as returned by :py:meth:`.Client.fetch_skeleton()`

        export_path:
            Optional. Write the SWC file to disk a the given location.

    Returns:
        ``str``
    """
    df = df.copy()
    df['node_type'] = 0
    df = df[['rowId', 'node_type', 'x', 'y', 'z', 'radius', 'link']]
    swc = "# "
    swc += df.to_csv(sep=' ', header=True, index=False)

    if export_path:
        with open(export_path, 'w') as f:
            f.write(swc)

    return swc


def heal_skeleton(skeleton_df, max_distance=np.inf, root_parent=None):
    """
    Attempt to repair a fragmented skeleton into a single connected component.

    Rather than a single tree, skeletons from neuprint sometimes
    consist of multiple fragments, i.e. multiple connected
    components.  That's due to artifacts in the underlying
    segmentation from which the skeletons were generated.
    In such skeletons, there will be multiple 'root' nodes
    (SWC rows where ``link == -1``).

    This function 'heals' a fragmented skeleton by joining its
    fragments into a single tree.

    First, each fragment is joined to every other fragment at
    their nearest points. The resulting graph has unnecessary
    edges, which are then removed by extracting the minimum
    spanning tree.  The MST is returned as the healed skeleton.

    Args:
        skeleton_df:
            DataFrame as returned by :py:meth:`.Client.fetch_skeleton()`

        max_distance:
            If a skeleton's fragments are very spatially distant, it may
            not be desirable to connect them with a new edge.
            This parameter specifies the maximum length of new edges
            introduced by the healing procedure.  If a skeleton fragment
            cannot be connected to the rest of the skeleton because it's
            too far away, the skeleton will remain fragmented.

        root_parent:
            Typically, root nodes point to a special ID, e.g. -1.
            If this ID is known, then 
    Returns:
        DataFrame, with ``link`` column updated with updated edges.
    """
    if max_distance is True:
        max_distance = np.inf

    if not max_distance:
        max_distance = 0.0

    if root_parent is None:
        root_parent = -1
    else:
        # Fast path to exit early if we can easily check the number of roots.
        num_roots = skeleton_df.eval(f'link == {root_parent}').sum()
        if num_roots == 1:
            # There's only one root and therefore only one component.
            # No healing necessary.
            return skeleton_df

    skeleton_df = skeleton_df.sort_values('rowId', ignore_index=True)
    g = skeleton_df_to_nx(skeleton_df, False, False)

    # Extract each fragment's rows and construct a KD-Tree
    Fragment = namedtuple('Fragment', ['frag_id', 'df', 'kd'])
    fragments = []
    for frag_id, cc in enumerate(nx.connected_components(g)):
        if len(cc) == len(skeleton_df):
            # There's only one component -- no healing necessary
            return skeleton_df
        df = skeleton_df.query('rowId in @cc')
        kd = cKDTree(df[[*'xyz']].values)
        fragments.append( Fragment(frag_id, df, kd) )

    # Sort from big-to-small, so the calculations below use a
    # KD tree for the larger point set in every fragment pair.
    fragments = sorted(fragments, key=lambda frag: -len(frag.df))

    # We could use the full graph and connect all
    # fragment pairs at their nearest neighbors,
    # but it's faster to treat each fragment as a
    # single node and run MST on that quotient graph,
    # which is tiny.
    frag_graph = nx.Graph()
    for frag_a, frag_b in combinations(fragments, 2):
        coords_b = frag_b.df[[*'xyz']].values
        distances, indexes = frag_a.kd.query(coords_b)

        index_b = np.argmin(distances)
        index_a = indexes[index_b]

        node_a = frag_a.df['rowId'].iloc[index_a]
        node_b = frag_b.df['rowId'].iloc[index_b]
        dist_ab = distances[index_b]

        # Add edge from one fragment to another,
        # but keep track of which fine-grained skeleton
        # nodes were used to calculate distance.
        frag_graph.add_edge( frag_a.frag_id, frag_b.frag_id,
                             node_a=node_a, node_b=node_b,
                             distance=dist_ab )

    # Compute inter-fragment MST edges
    frag_edges = nx.minimum_spanning_edges(frag_graph, weight='distance', data=True)

    # For each inter-fragment edge, add the corresponding
    # fine-grained edge between skeleton nodes in the original graph.
    omit_edges = []
    for _u, _v, d in frag_edges:
        g.add_edge(d['node_a'], d['node_b'])
        if d['distance'] > max_distance:
            omit_edges.append((d['node_a'], d['node_b']))

    # Traverse in depth-first order to compute edges for final tree
    root = skeleton_df['rowId'].iloc[0]

    # Replace 'link' (parent) column using MST edges
    _reorient_skeleton(skeleton_df, root, root_parent, g=g)
    assert (skeleton_df['link'] == root_parent).sum() == 1
    assert skeleton_df['link'].iloc[0] == root_parent

    # Delete edges that violated max_distance
    for a,b in omit_edges:
        q = '(rowId == @a and link == @b) or (rowId == @b and link == @a)'
        idx = skeleton_df.query(q).index
        skeleton_df.loc[idx, 'link'] = root_parent

    return skeleton_df


def _reorient_skeleton(skeleton_df, root, root_parent=-1, g=None):
    """
    Replace the 'link' column in each row of the skeleton dataframe
    so that its parent corresponds to a depth-first traversal from
    the given root node.

    Args:
        skeleton_df:
            A skeleton dataframe

        root:
            A rowId to use as the new root node

        g:
            Optional. A nx.Graph representation of the skeleton

    Works in-place.
    """
    g = g or skeleton_df_to_nx(skeleton_df, False, False)
    assert isinstance(g, nx.Graph) and not isinstance(g, nx.DiGraph), \
        "skeleton graph must be undirected"

    edges = list(nx.dfs_edges(g, source=root))

    # If the graph has more than one connected component,
    # the remaining components have arbitrary roots
    if len(edges) != len(g.edges):
        for cc in nx.connected_components(g):
            if root not in cc:
                edges += list(nx.dfs_edges(g, source=cc.pop()))

    edges = pd.DataFrame(edges, columns=['link', 'rowId'])  # parent, child
    edges = edges.set_index('rowId')['link']

    # Replace 'link' (parent) column using DFS edges
    skeleton_df['link'] = skeleton_df['rowId'].map(edges).fillna(root_parent).astype(int)


def reorient_skeleton(skeleton_df, rowId=None, xyz=None, use_max_radius=False):
    """
    Change the root node of a skeleton.

    In general, the root node of the skeletons stored in neuprint is
    not particularly significant, so the directionality of the nodes
    (parent to child or vice-versa) on any given neuron branch is arbitrary.

    This function allows you to pick a different root node and reorient
    the tree with respect to that node. Replaces the 'link' column in
    each row of the skeleton dataframe so that its parent corresponds
    to a depth-first traversal from the new root node.

    You can specify the new root node either by its row, or by a coordinate
    (the closest node to that coordinate will be selected) or by size
    (the largest node will be selected).

    Works in-place.  Only the 'link' column is changed.

    If the given skeleton has more than one connected component (and thus
    more than one root node), the orientation of the edges in other components
    will be arbitrary.

    Args:
        skeleton_df:
            A skeleton dataframe, e.g. as returned by `py:func:fetch_skeleton(..., heal=True)`

        rowId:
            A rowId to use as the new root node

        xyz:
            If given, chooses the node closest to the given coordinate as the new root node.

        use_max_radius:
            If True, choose the largest node (by radius) to use as the new root node.
    """
    assert rowId != 0, \
        "rowId is never 0 in NeuTu skeletons"

    assert bool(rowId) + (xyz is not None) + use_max_radius == 1, \
        "Select either a rowId to use as the new root, or a coordinate, or use_max_radius=True"

    if xyz is not None:
        # Find closest node to the given coordinate
        distances = np.linalg.norm(skeleton_df[[*'xyz']] - xyz, axis=1)
        rowId = skeleton_df['rowId'].iloc[np.argmin(distances)]
    elif use_max_radius:
        # Find the node with the largest radius
        idx = skeleton_df['radius'].idxmax()
        rowId = skeleton_df.loc[idx, 'rowId']

    assert rowId is not None, "You must specify a new root node"

    _reorient_skeleton(skeleton_df, rowId)


def skeleton_segments(skeleton_df):
    """
    Compute a table of skeleton segments.

    A skeleton dataframe is a table of nodes (points) and their parent nodes.

    This function computes a table of segments, where each row lists both the
    child and parent point, along with some attributes describing the segment:
    length, average radius, and segment volume.
    """
    assert isinstance(skeleton_df, pd.DataFrame)

    segment_df = skeleton_df.merge(skeleton_df[['rowId', 'link', *'xyz', 'radius']],
                                   'inner',
                                   left_on='link',
                                   right_on='rowId',
                                   suffixes=['', '_parent'])

    child_points = segment_df[[*'xyz']].values
    parent_points = segment_df[['x_parent', 'y_parent', 'z_parent']].values
    segment_df['length'] = np.linalg.norm(child_points - parent_points, axis=1)
    segment_df['avg_radius'] = segment_df.eval('(radius + radius_parent) / 2')

    # Volume of a truncated cone:
    # V = π * h * (R² * r² + R*r) / 3
    PI = np.pi
    e = '@PI * length * (radius_parent**2 + radius**2 + radius*radius_parent) / 3'
    segment_df['volume'] = segment_df.eval(e)

    return segment_df


def attach_synapses_to_skeleton(skeleton_df, synapses_df):
    """
    Attach a neuron's synapses to its skeleton as new skeleton nodes.

    Synapses are attached to their nearest skeleton node (in euclidean terms).
    Note that skeleton nodes are sometimes somewhat far apart, so the nearest node
    to a given synapse may not be close to the nearest point on the nearest line segment.

    Args:
        skeleton_df:
            A DataFrame containing a neuron skeleton,
            as produced by ``fetch_skeleton()``

        synapses_df:
            A DataFrame of synapse points, as produced by ``fetch_synapses()``.

    Returns:
        DataFrame
        Rows are appended to the skeleton (one per synapse), and a new column is
        added to distinguish between nodes which belonged to the original skeleton
        ('neurite') and those which were added for synapses ('pre' or 'post').

    Example:

        .. code-block:: ipython

                In [4]: from neuprint import fetch_skeleton, fetch_synapses, attach_synapses_to_skeleton
                   ...: body = 1136399017
                   ...: skeleton = fetch_skeleton(body, heal=True)
                   ...: synapses = fetch_synapses(body)
                   ...: attach_synapses_to_skeleton(skeleton, synapses)

                Out[4]:
                    rowId             x             y             z     radius  link structure
                0         1  12798.000000  30268.000000  15812.000000  21.000000    -1   neurite
                1         2  12746.700195  30370.699219  15788.700195  55.464100     1   neurite
                2         3  12727.200195  30411.199219  15767.599609  68.081200     2   neurite
                3         4  12705.299805  30475.599609  15716.400391  58.952702     3   neurite
                4         5  12687.400391  30499.500000  15692.500000  50.619999     4   neurite
                ...     ...           ...           ...           ...        ...   ...       ...
                2032   2033  12073.000000  32575.000000  14386.000000   0.000000   651      post
                2033   2034  10072.000000  32572.000000  14464.000000   0.000000   685      post
                2034   2035  10647.000000  31797.000000  14057.000000   0.000000   760      post
                2035   2036  11203.000000  30673.000000  14839.000000   0.000000  1086      post
                2036   2037  11116.000000  30613.000000  14707.000000   0.000000  1068      post

                [2037 rows x 7 columns]
    """
    skeleton_df = skeleton_df.copy(deep=False)
    synapses_df = synapses_df.copy(deep=False)

    kd = cKDTree(skeleton_df[[*'xyz']].values)
    _, indexes = kd.query(synapses_df[[*'xyz']].values)
    synapses_df['link'] = skeleton_df.loc[indexes, 'rowId'].values
    skeleton_df['structure'] = 'neurite'

    synapses_df['rowId'] = synapses_df.index + skeleton_df['rowId'].max() + 1
    synapses_df['structure'] = synapses_df['type']
    synapses_df['radius'] = 0.0

    relevant_cols = ['rowId', *'xyz', 'radius', 'link', 'structure']
    combined = pd.concat((skeleton_df, synapses_df[relevant_cols]), ignore_index=True)
    combined['structure'] = pd.Categorical(combined['structure'])
    return combined



from .neuroncriteria import NeuronCriteria, SegmentCriteria
from .mitocriteria import MitoCriteria
from .synapsecriteria import SynapseCriteria
from .general import fetch_custom, fetch_meta
from .rois import fetch_all_rois, fetch_primary_rois,fetch_roi_hierarchy
from .neurons import fetch_neurons, fetch_custom_neurons
from .connectivity import (fetch_simple_connections, fetch_adjacencies, fetch_traced_adjacencies,
                           fetch_common_connectivity, fetch_shortest_paths)
from .synapses import fetch_synapses, fetch_mean_synapses, fetch_synapse_connections
from .mito import fetch_mitochondria, fetch_synapses_and_closest_mitochondria, fetch_connection_mitochondria
from .recon import fetch_output_completeness, fetch_downstream_orphan_tasks

# Change the __module__ of each function to make it look like it was defined in this file.
# This hackery is to make the Sphinx autosummary and autodoc play nicely together.
for k, v in dict(locals()).items():
    if k.startswith('fetch'):
        v.__module__ = 'neuprint.queries'
del k, v



import re
import copy
import inspect
import functools
import collections.abc
from itertools import chain
from textwrap import indent, dedent
from collections.abc import Iterable, Collection

import numpy as np
import pandas as pd

from ..utils import make_args_iterable, IsNull, NotNull
from ..client import inject_client

NoneType = type(None)


def neuroncriteria_args(*argnames):
    """
    Returns a decorator.
    For the given argument names, the decorator converts the
    arguments into NeuronCriteria objects via ``copy_as_neuroncriteria()``.

    If the decorated function also accepts a 'client' argument,
    that argument is used to initialize the NeuronCriteria.
    """
    def decorator(f):

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            callargs = inspect.getcallargs(f, *args, **kwargs)
            for name in argnames:
                callargs[name] = copy_as_neuroncriteria(callargs[name], callargs.get('client', None))
            return f(**callargs)

        wrapper.__signature__ = inspect.signature(f)
        return wrapper

    return decorator


def copy_as_neuroncriteria(obj, client=None):
    """
    If the given argument is a NeuronCriteria object, copy it.
    Otherwise, attempt to construct a NeuronCriteria object,
    using the argument as either the bodyId or the type AND instance.

    Rules:

        NC -> copy(NC)

        None -> NC()

        int -> NC(bodyId)
        [int,...] -> NC(bodyId)
        DataFrame['bodyId'] -> NC(bodyId)

        str -> NC(type, instance)
        [str, ...] -> NC(type, instance)

        [] -> Error
        [None] -> Error
        Anything else -> Error

    """
    if isinstance(obj, pd.DataFrame):
        assert 'bodyId' in obj.columns, \
            'If passing a DataFrame as NeuronCriteria, it must have "bodyId" column'
        return NeuronCriteria(bodyId=obj['bodyId'].values, client=client)

    if obj is None:
        return NeuronCriteria(client=client)

    if isinstance(obj, NeuronCriteria):
        return copy.copy(obj)

    if not isinstance(obj, Collection) or isinstance(obj, str):
        if isinstance(obj, str):
            return NeuronCriteria(type=obj, instance=obj, client=client)

        if np.issubdtype(type(obj), np.integer):
            return NeuronCriteria(bodyId=obj, client=client)

        raise RuntimeError(f"Can't auto-construct a NeuronCriteria from {obj}.  Please explicitly create one.")
    else:
        if len(obj) == 0:
            raise RuntimeError(f"Can't auto-construct a NeuronCriteria from {obj}.  Please explicitly create one.")

        if len(obj) == 1 and obj[0] is None:
            raise RuntimeError(f"Can't auto-construct a NeuronCriteria from {obj}.  Please explicitly create one.")

        if isinstance(obj, np.ndarray) and np.issubdtype(obj.dtype, np.integer):
            return NeuronCriteria(bodyId=obj, client=client)

        item = [*filter(lambda item: item is not None, obj)][0]
        if np.issubdtype(type(item), np.integer):
            return NeuronCriteria(bodyId=obj, client=client)

        if isinstance(item, str):
            return NeuronCriteria(type=obj, instance=obj, client=client)

        raise RuntimeError(f"Can't auto-construct a NeuronCriteria from {obj}.  Please explicitly create one.")


class NeuronCriteria:
    """
    Neuron selection criteria.

    Specifies which fields to filter by when searching for a Neuron (or Segment).
    This class does not send queries itself, but you use it to specify search
    criteria for various query functions.

    Note:
        For simple queries involving only particular bodyId(s) or type(s)/instance(s),
        you can usually just pass the ``bodyId`` or ``type`` to the query function,
        without constructing a full ``NeuronCriteria``.

        .. code-block:: python

            from neuprint import fetch_neurons, NeuronCriteria as NC

            # Equivalent
            neuron_df, conn_df = fetch_neurons(NC(bodyId=329566174))
            neuron_df, conn_df = fetch_neurons(329566174)

            # Equivalent
            # (Criteria is satisfied if either type or instance matches.)
            neuron_df, conn_df = fetch_neurons(NC(type="OA-VPM3", instance="OA-VPM3"))
            neuron_df, conn_df = fetch_neurons("OA-VPM3")
    """

    @inject_client
    @make_args_iterable(['bodyId', 'instance', 'type', 'cellBodyFiber',
                         'status', 'statusLabel', 'rois', 'inputRois', 'outputRois',
                         'hemilineage', 'class_', 'exitNerve'])
    def __init__( self, matchvar='n', *,
                  bodyId=None, type=None, instance=None, regex='guess',
                  class_=None, somaSide=None, exitNerve=None, hemilineage=None,
                  cellBodyFiber=None,
                  status=None, statusLabel=None, cropped=None,
                  min_pre=0, min_post=0,
                  rois=None, inputRois=None, outputRois=None, min_roi_inputs=1, min_roi_outputs=1,
                  label=None, roi_req='all',
                  soma=None,
                  client=None ):
        """
        Except for ``matchvar``, all parameters must be passed as keyword arguments.

        .. note::

            **Options for specifying ROI criteria**

            The ``rois`` argument merely matches neurons that intersect the given ROIs at all
            (without distinguishing between inputs and outputs).

            The ``inputRois`` and ``outputRois`` arguments allow you to put requirements
            on whether or not neurons have inputs or outputs in the listed ROIs.
            It results a more expensive query, but its more powerful.
            It also enables you to require a minimum number of connections in the given
            ``inputRois`` or ``outputRois`` using the ``min_roi_inputs`` and ``min_roi_outputs``
            criteria.

            In either case, use use ``roi_req`` to specify whether a neuron must match just
            one (``any``) of the listed ROIs, or ``all`` of them.

        .. note::

            **Matching against missing values (NULL)**

            To search for neurons which are missing given property entirely,
            you can use a list containing ``None``, or the special value ``neuprint.IsNull``.
            For example, to find neurons with no `type`, use ``type=[None]`` or ``type=IsNull``.

            **Matching against any value (NOT NULL)**

            To search for any non-null value, you can use ``neuprint.NotNull``. For
            example, to find neurons that have a type (no matter what the
            type is), use ``type=NotNull``.

        Args:
            matchvar (str):
                An arbitrary cypher variable name to use when this
                ``NeuronCriteria`` is used to construct cypher queries.
                To help catch errors (such as accidentally passing a ``type`` or
                ``instance`` name in the wrong argument position), we require that
                ``matchvar`` begin with a lowercase letter.

            bodyId (int or list of ints):
                List of bodyId values.

            type (str or list of str):
                Cell type.  Matches depend on the the ``regex`` argument.
                To search for neurons with no type at all, use ``type=[None]``.
                If both ``type`` and ``instance`` criteria are supplied, any neuron that
                matches EITHER criteria will match the overall criteria.

            instance (str or list of str):
                Cell instance (specific cell name).  Matches depend on the the ``regex`` argument.
                To search for neurons with no instance at all, use ``instance=[None]``.
                If both ``type`` and ``instance`` criteria are supplied, any neuron that
                matches EITHER criteria will match the overall criteria.

            regex (bool):
                If ``True``, the ``instance`` and ``type`` arguments will be interpreted as
                regular expressions, rather than exact match strings.
                If ``False``, only exact matches will be found.
                By default, the matching method will be automatically chosen by inspecting the
                ``type`` and ``instance`` strings.  If they contain regex-like characters,
                then we assume you intend regex matching. (You can see which method was chosen by
                checking the ``regex`` field after the ``NeuronCriteria`` is constructed.)

            class_ (str or list of str):
                Matches for the neuron ``class`` field.  To search for neurons
                with no class at all, use ``class_=[None]``.

            somaSide ('RHS' or 'LHS' or None):
                Matches for the neuron ``somaSide`` field.

            exitNerve (str or list of str):
                Matches for the neuron ``exitNerve`` field.  To search for neurons
                with no exit nerve at all, use ``exitNerve=[None]``.

            hemilineage (str or list of str):
                Matches for the neuron ``hemilineage`` field.  To search for neurons
                with no hemilineage at all, use ``hemilineage=[None]``.

            cellBodyFiber (str or list of str):
                Matches for the neuron ``cellBodyFiber`` field.  To search for neurons
                with no CBF at all, use ``cellBodyFiber=[None]``.

            status (str or list of str):
                Matches for the neuron ``status`` field.  To search for neurons with no status
                at all, use ``status=[None]``.

            statusLabel (str or list of str):
                Matches for the neuron ``statusLabel`` field. ``statusLabel`` is
                typically more finegrained than ``status``. To search for neurons
                with no status at all, use ``statusLabel=[None]``.

            cropped (bool):
                If given, restrict results to neurons that are cropped or not.

            min_pre (int):
                Exclude neurons that don't have at least this many t-bars (outputs) overall,
                regardless of how many t-bars exist in any particular ROI.

            min_post (int):
                Exclude neurons that don't have at least this many PSDs (inputs) overall,
                regardless of how many PSDs exist in any particular ROI.

            rois (str or list of str):
                ROIs that merely intersect the neuron, without specifying whether
                they're intersected by input or output synapses.
                If not provided, will be auto-set from ``inputRois`` and ``outputRois``.

            inputRois (str or list of str):
                Only Neurons which have inputs in EVERY one of the given ROIs will be matched.
                ``regex`` does not apply to this parameter.

            outputRois (str or list of str):
                Only Neurons which have outputs in EVERY one of the given ROIs will be matched.
                ``regex`` does not apply to this parameter.

            min_roi_inputs (int):
                How many input (post) synapses a neuron must have in each ROI to satisfy the
                ``inputRois`` criteria.  Can only be used if you provided ``inputRois``.

            min_roi_outputs (int):
                How many output (pre) synapses a neuron must have in each ROI to satisfy the
                ``outputRois`` criteria.   Can only be used if you provided ``outputRois``.

            label (Either ``'Neuron'`` or ``'Segment'``):
                Which node label to match with.
                (In neuprint, all ``Neuron`` nodes are also ``Segment`` nodes.)
                By default, ``'Neuron'`` is used, unless you provided a non-empty ``bodyId`` list.
                In that case, ``'Segment'`` is the default. (It's assumed you're really interested
                in the bodies you explicitly listed, whether or not they have the ``'Neuron'`` label.)

            roi_req (Either ``'any'`` or ``'all'``):
                Whether a neuron must intersect all of the listed input/output ROIs, or any of the listed input/output ROIs.
                When using 'any', each neuron must still match at least one input AND at least one output ROI.

            soma (Either ``True``, ``False``, or ``None``)
                If ``True``, only return neurons with a ``somaLocation``.
                If ``False``, return neurons without a ``somaLocation``.

            client (:py:class:`neuprint.client.Client`):
                Used to validate ROI names.
                If not provided, the global default ``Client`` will be used.

        """
        # Validate that matchvar in various ways, to catch errors in which
        # the user has passed a bodyId or type, etc. in the wrong position.
        assert isinstance(matchvar, str), \
            (f"Bad matchvar argument (should be str): {matchvar}. "
             "Did you mean to pass this as bodyId, type, or instance name?")
        assert matchvar, "matchvar cannot be an empty string"
        assert re.match('^[a-z].*$', matchvar), \
            (f"matchvar must begin with a lowercase letter, not '{matchvar}'. "
             "Did you mean to pass this as a type or instance name?")
        assert re.match('^[a-zA-Z0-9]+$', matchvar), \
            (f"matchvar contains invalid characters: '{matchvar}'. "
             "Did you mean to pass this as a type or instance?")

        assert len(bodyId) == 0 or np.issubdtype(np.asarray(bodyId).dtype, np.integer), \
            "bodyId should be an integer or list of integers"

        if not label:
            if len(bodyId) == 0:
                label = 'Neuron'
            else:
                label = 'Segment'
        assert label in ('Neuron', 'Segment'), f"Invalid label: {label}"

        assert regex in (True, False, 'guess')

        for i in instance:
            assert isinstance(i, (str, NoneType)) or i in (IsNull, NotNull), \
                f'instance should be a string, IsNull, NotNull or None, got {i}'

        for t in type:
            assert isinstance(t, (str, NoneType)) or t in (IsNull, NotNull), \
                f'type should be a string, IsNull, NotNull or None, got {t}'

        if regex == 'guess':
            rgx = re.compile(r'[\\\.\?\[\]\+\^\$\*]')
            instance_is_regex = False
            for i in instance:
                instance_is_regex |= isinstance(i, str) and bool(rgx.search(i or ''))

            type_is_regex = False
            for t in type:
                type_is_regex |= isinstance(t, str) and bool(rgx.search(t or ''))

            regex = type_is_regex or instance_is_regex

        assert roi_req in ('any', 'all')

        assert min_roi_inputs <= 1 or inputRois, \
            "Can't stipulate min_roi_inputs without a list of inputRois"
        assert min_roi_outputs <= 1 or outputRois, \
            "Can't stipulate min_roi_outputs without a list of outputRois"

        assert soma in (True, False, None), \
            f"soma must be True, False or None, not {soma}"

        assert somaSide in ("RHS", "LHS", None), \
            f"somaSide must be 'LHS', 'RHS' or None, not {somaSide}"

        # If the user provided both intersecting rois and input/output rois,
        # force them to make the intersecting set a superset of the others.
        rois = {*rois}
        inputRois = {*inputRois}
        outputRois = {*outputRois}
        assert not rois or rois >= {*inputRois}, "Queried intersecting rois must be a superset of the inputRois"
        assert not rois or rois >= {*outputRois}, "Queried intersecting rois must be a superset of the outputRois"

        # Make sure intersecting is a superset of inputRois and outputRois
        rois |= {*inputRois, *outputRois}

        # Verify ROI names against known ROIs.
        neuprint_rois = {*client.all_rois}
        unknown_input_rois = inputRois - neuprint_rois
        if unknown_input_rois:
            raise RuntimeError(f"Unrecognized input ROIs: {unknown_input_rois}")

        unknown_output_rois = outputRois - neuprint_rois
        if unknown_output_rois:
            raise RuntimeError(f"Unrecognized output ROIs: {unknown_output_rois}")

        unknown_generic_rois = rois - neuprint_rois
        if unknown_generic_rois:
            raise RuntimeError(f"Unrecognized output ROIs: {unknown_generic_rois}")

        self.matchvar = matchvar
        self.bodyId = bodyId
        self.instance = instance
        self.type = type
        self.cellBodyFiber = cellBodyFiber
        self.status = status
        self.statusLabel = statusLabel
        self.cropped = cropped
        self.min_pre = min_pre
        self.min_post = min_post
        self.rois = rois
        self.inputRois = inputRois
        self.outputRois = outputRois
        self.min_roi_inputs = min_roi_inputs
        self.min_roi_outputs = min_roi_outputs
        self.regex = regex
        self.label = label
        self.roi_req = roi_req
        self.soma = soma
        self.class_ = class_
        self.somaSide = somaSide
        self.exitNerve = exitNerve
        self.hemilineage = hemilineage

        self.list_props = ['bodyId', 'status', 'statusLabel', 'cellBodyFiber',
                           'hemilineage', 'exitNerve', 'class_']
        self.list_props_regex = ['type', 'instance']

    def __eq__(self, value):
        """
        Implement comparison between criteria.
        Note: 'matchvar' is not considered during the comparison.
        """
        if not isinstance(value, NeuronCriteria):
            return NotImplemented

        # Return True if it's the exact same object
        if self is value:
            return True

        # Compare attributes one by one
        # But don't count 'matchvar' as a parameter'.
        params = [#'matchvar',
                 'bodyId', 'instance', 'type', 'status', 'statusLabel',
                 'cropped', 'min_pre', 'min_post', 'rois', 'inputRois',
                 'outputRois', 'min_roi_inputs', 'min_roi_outputs',
                 'regex', 'label', 'roi_req', 'soma']

        for at in params:
            me = getattr(self, at)
            other = getattr(value, at)

            # If not the same type, return False
            if type(me) != type(other):
                return False

            # If iterable (e.g. ROIs or body IDs) we don't care about order
            if isinstance(me, Iterable):
                if not all([v in other for v in me]):
                    return False
            elif me != other:
                return False
        # If all comparisons have passed, return True
        return True

    def __repr__(self):
        # Show all non-default constructor args
        s = f'NeuronCriteria("{self.matchvar}"'

        if len(self.bodyId):
            s += f", bodyId={list(self.bodyId)}"

        if len(self.instance) == 1:
            s += f', instance="{self.instance[0]}"'
        elif len(self.instance) > 1:
            s += f", instance={list(self.instance)}"

        if len(self.type) == 1:
            s += f', type="{self.type[0]}"'
        elif len(self.instance) > 1:
            s += f", type={list(self.type)}"

        if len(self.type) or len(self.instance):
            s += f", regex={self.regex}"

        if len(self.cellBodyFiber) == 1:
            s += f', cellBodyFiber="{self.cellBodyFiber[0]}"'
        elif len(self.instance) > 1:
            s += f", cellBodyFiber={list(self.cellBodyFiber)}"

        if len(self.status) == 1:
            s += f', status="{self.status[0]}"'
        elif len(self.instance) > 1:
            s += f", status={list(self.status)}"

        if len(self.statusLabel) == 1:
            s += f', statusLabel="{self.statusLabel[0]}"'
        elif len(self.instance) > 1:
            s += f", statusLabel={list(self.statusLabel)}"

        if self.cropped is not None:
            s += f", cropped={self.cropped}"

        if self.min_pre != 0:
            s += f", min_pre={self.min_pre}"

        if self.min_post != 0:
            s += f", min_post={self.min_post}"

        if self.rois:
            s += f", rois={list(self.rois)}"

        if self.inputRois:
            s += f", inputRois={list(self.inputRois)}"

        if self.outputRois:
            s += f", outputRois={list(self.outputRois)}"

        if self.min_roi_inputs != 1:
            s += f", min_roi_inputs={self.min_roi_inputs}"

        if self.min_roi_outputs != 1:
            s += f", min_roi_outputs={self.min_roi_outputs}"

        if self.label != 'Neuron':
            s += f', label="{self.label}"'

        if self.roi_req != 'all':
            s += f', roi_req="{self.roi_req}"'

        if self.soma is not None:
            s += f', soma="{self.soma}"'

        s += ')'

        return s

    MAX_LITERAL_LENGTH = 3
    assert MAX_LITERAL_LENGTH >= 3, \
        ("The logic in where_expr() assumes valuevars "
         "have length 3 (assuming one could be None).")

    def global_vars(self):
        exprs = {}

        if self.regex:
            list_props = self.list_props
        else:
            # No regex: Treat instance and type as ordinary strings
            list_props = self.list_props + self.list_props_regex

        for key in list_props:
            values = getattr(self, key)
            if len(values) > self.MAX_LITERAL_LENGTH:
                if key.endswith('_'):
                    key = key[:-1]
                values = [*filter(lambda s: s is not None, values)]
                var = f"{self.matchvar}_search_{key}"
                exprs[var] = (f"{[*values]} as {var}")

        return exprs

    def global_with(self, *vars, prefix=0):
        if isinstance(prefix, int):
            prefix = ' '*prefix

        if vars:
            carry_forward = [', '.join(vars)]
        else:
            carry_forward = []

        full_list = ',\n     '.join([*carry_forward, *self.global_vars().values()])
        if full_list:
            return indent('WITH ' + full_list, prefix)[len(prefix):]
        return ""

    def basic_exprs(self):
        """
        Return the list of expressions that correspond
        to the members in this NeuronCriteria object.
        They're intended be combined (via 'AND') in
        the WHERE clause of a cypher query.
        """
        exprs = [self.bodyId_expr(), self.typeinst_expr(), self.cbf_expr(),
                 self.status_expr(), self.statusLabel_expr(),
                 self.cropped_expr(), self.rois_expr(), self.pre_expr(), self.post_expr(),
                 self.soma_expr(), self.hemilineage_expr(), self.class_expr(),
                 self.exitNerve_expr(), self.somaSide_expr()]
        exprs = [*filter(None, exprs)]
        return exprs

    def _value_list_expr(self, key, value, regex=False):
        """
        Match key against a list of values. E.g. "bodyId in [1234, 5678]".
        """
        valuevar = None
        if not regex and len(value) > self.MAX_LITERAL_LENGTH:
            valuevar = f"{self.matchvar}_search_{key}"
        return where_expr(key, value, regex, self.matchvar, valuevar)

    def _single_value_expr(self, key, value):
        """
        Match against key/value:
            - True: key must exist
            - False: key must not exist
            - str: key must have given value
        """
        if value is None:
            return ""
        if not isinstance(value, bool):
            return f"{self.matchvar}.{key} = '{value}'"
        elif value:
            return f"{self.matchvar}.{key} IS NOT NULL"
        else:
            return f"{self.matchvar}.{key} IS NULL"

    def _tag_expr(self, key, value):
        """
        Match against tag, e.g. `.cropped`.
        Non-existing tags are counted as False.
        """
        if value is None:
            return ""

        if value:
            return f"{self.matchvar}.{key}"
        else:
            # Not all neurons might actually have the flag (e.g. `.cropped`),
            # so simply checking for False values isn't enough.
            # Must check exists().
            return f"(NOT {self.matchvar}.{key} OR NOT exists({self.matchvar}.{key}))"

    def _logic_tag_expr(self, tags, logic):
        """
        Match against logic list of tags, e.g. `.LH(R)` AND `.AL(R)`.
        """
        assert logic in ('AND', 'OR'), '`logic` must be either AND or OR'
        if len(tags) == 0:
            return ""

        tags = sorted(tags)
        return "(" + f" {logic} ".join(f"{self.matchvar}.`{v}`" for v in tags) + ")"

    def _gt_eq_expr(self, key, value):
        """
        Match against key/value being greater or equal.
        """
        if value:
            return f"{self.matchvar}.{key} >= {value}"
        else:
            return ""

    def typeinst_expr(self):
        """
        Unlike all other fields, type and instance OR'd together.
        Either match satisfies the criteria.
        """
        t = self.type_expr()
        i = self.instance_expr()

        if t and i:
            return f"({t} OR {i})"
        if t:
            return t
        if i:
            return i
        return ""

    def bodyId_expr(self):
        return self._value_list_expr('bodyId', self.bodyId, False)

    def instance_expr(self):
        return self._value_list_expr('instance', self.instance, self.regex)

    def type_expr(self):
        return self._value_list_expr('type', self.type, self.regex)

    def cbf_expr(self):
        return self._value_list_expr('cellBodyFiber', self.cellBodyFiber, False)

    def status_expr(self):
        return self._value_list_expr('status', self.status, False)

    def statusLabel_expr(self):
        return self._value_list_expr('statusLabel', self.statusLabel, False)

    def hemilineage_expr(self):
        return self._value_list_expr('hemilineage', self.hemilineage, False)

    def exitNerve_expr(self):
        return self._value_list_expr('exitNerve', self.exitNerve, False)

    def class_expr(self):
        return self._value_list_expr('class', self.class_, False)

    def cropped_expr(self):
        return self._tag_expr('cropped', self.cropped)

    def rois_expr(self):
        return self._logic_tag_expr(self.rois,
                               {'any': 'OR', 'all': 'AND'}[self.roi_req])

    def pre_expr(self):
        return self._gt_eq_expr('pre', self.min_pre)

    def post_expr(self):
        return self._gt_eq_expr('post', self.min_post)

    def soma_expr(self):
        return self._single_value_expr('somaLocation', self.soma)

    def somaSide_expr(self):
        return self._single_value_expr('somaSide', self.somaSide)

    def all_conditions(self, *vars, prefix=0, comments=True):
        if isinstance(prefix, int):
            prefix = ' '*prefix

        vars = {*vars} | {self.matchvar, *self.global_vars().keys()}
        vars = (*vars,)

        basic_cond = self.basic_conditions(0, comments)
        if basic_cond:
            basic_cond = f"WHERE \n{basic_cond}"
            basic_cond = indent(basic_cond, '  ')[2:]

        roi_cond = self.directed_rois_condition(*vars, comments=comments)

        if roi_cond:
            combined = basic_cond + "\n\n" + roi_cond
        else:
            combined = basic_cond

        return indent(combined, prefix)[len(prefix):]

    @classmethod
    def combined_global_with(cls, neuron_conditions, vars=[], prefix=0):
        if isinstance(prefix, int):
            prefix = ' '*prefix

        if vars:
            carry_forward = [', '.join(vars)]
        else:
            carry_forward = []

        all_globals = chain(*(nc.global_vars().values() for nc in neuron_conditions))
        full_list = ',\n     '.join([*carry_forward, *all_globals])

        if full_list:
            return indent('WITH ' + full_list, prefix)[len(prefix):]
        return ""

    @classmethod
    def combined_conditions(cls, neuron_conditions, vars=[], prefix=0, comments=True):
        """
        Combine the conditions from multiple NeuronCriteria into a single string,
        putting the "cheap" conditions first and the "expensive" conditions last.
        (That is, basic conditions first and the directed ROI conditions last.)
        """
        if isinstance(prefix, int):
            prefix = ' '*prefix

        vars = {*vars}
        for nc in neuron_conditions:
            vars = vars | {nc.matchvar, *nc.global_vars().keys()}
        vars = (*vars,)

        basic_cond = [nc.basic_conditions(0, comments) for nc in neuron_conditions]
        basic_cond = [*filter(None, basic_cond)]
        if not basic_cond:
            return ""

        if basic_cond:
            basic_cond = '\nAND\n'.join(basic_cond)
            basic_cond = indent(basic_cond, ' '*2)
            basic_cond = f"WHERE \n{basic_cond}"

        combined = basic_cond

        roi_conds = [nc.directed_rois_condition(*vars, comments=comments) for nc in neuron_conditions]
        roi_conds = [*filter(None, roi_conds)]
        if roi_conds:
            roi_conds = '\n\n'.join(roi_conds)
            combined = basic_cond + "\n\n" + roi_conds

        return indent(combined, prefix)[len(prefix):]

    def basic_conditions(self, prefix=0, comments=True):
        """
        Construct a WHERE clause based on the basic conditions
        in this criteria (i.e. everything except for the "directed ROI" conditions.)
        """
        exprs = self.basic_exprs()
        if not exprs:
            return ""

        if isinstance(prefix, int):
            prefix = prefix*' '

        # Build WHERE clause by combining exprs for each field
        clauses = ""
        if comments:
            clauses += f"// -- Basic conditions for segment '{self.matchvar}' --\n"

        clauses += f"\nAND ".join(exprs)

        return indent(clauses, prefix)[len(prefix):]

    def directed_rois_condition(self, *vars, prefix=0, comments=True):
        """
        Construct the ```WITH...WHERE``` statements that apply the "directed ROI"
        conditions specified by this criteria's ``inputRois`` and ``outputRois``
        members.

        These conditions are expensive to evaluate, so it's usually a good
        idea to position them LAST in your cypher query, once the result set
        has already been narrowed down by eariler filters.
        """
        if not self.inputRois and not self.outputRois:
            return ""

        if isinstance(prefix, int):
            prefix = prefix*' '

        if len(self.inputRois) == 0:
            min_input_matches = 0
        elif self.roi_req == 'any':
            min_input_matches = 1
        elif self.roi_req == 'all':
            min_input_matches = 'size(inputRois)'
        else:
            assert False

        if len(self.outputRois) == 0:
            min_output_matches = 0
        elif self.roi_req == 'any':
            min_output_matches = 1
        elif self.roi_req == 'all':
            min_output_matches = 'size(outputRois)'
        else:
            assert False

        if vars:
            assert self.matchvar in vars, "Pass all match vars, including the one that belongs to this criteria"
            vars = ', '.join(vars)
        else:
            vars = self.matchvar

        conditions = dedent(f"""\
            // -- Directed ROI conditions for segment '{self.matchvar}' --
            WITH {vars},
                 {[*self.inputRois]} as inputRois,
                 {[*self.outputRois]} as outputRois,
                 apoc.convert.fromJsonMap({self.matchvar}.roiInfo) as roiInfo

            // Check input ROIs (segment '{self.matchvar}')
            UNWIND keys(roiInfo) as roi
            WITH {vars}, roi, roiInfo, inputRois, outputRois, roiInfo[roi]['post'] as roi_post
            ORDER BY roi
            // No filter if no input ROIs were specified, otherwise select the ones that meet the reqs
            WHERE {min_input_matches} = 0 OR (roi in inputRois AND roi_post >= {self.min_roi_inputs})
            WITH {vars}, roiInfo, inputRois, outputRois, collect(roi) as matchingInputRois, size(collect(roi)) as numMatchingInputRois
            WHERE numMatchingInputRois >= {min_input_matches}

            // Check output ROIs (segment '{self.matchvar}')
            UNWIND keys(roiInfo) as roi
            WITH {vars}, roi, roiInfo, inputRois, outputRois, matchingInputRois, roiInfo[roi]['pre'] as roi_pre
            ORDER BY roi
            // No filter if no output ROIs were specified, otherwise select the ones that meet the reqs
            WHERE {min_output_matches} = 0 OR (roi in outputRois AND roi_pre >= {self.min_roi_outputs})
            WITH {vars}, inputRois, outputRois, matchingInputRois, collect(roi) as matchingOutputRois, size(collect(roi)) as numMatchingOutputRois
            WHERE numMatchingOutputRois >= {min_output_matches}
            """)
        #RETURN n, matchingInputRois, matchingOutputRois

        if not comments:
            conditions = '\n'.join(filter(lambda s: '//' not in s, conditions.split('\n')))

        return indent(conditions, prefix)[len(prefix):]


#: Same as ``NeuronCriteria``.  This name is deprecated, but kept for backwards compatibility.
SegmentCriteria = NeuronCriteria


def where_expr(field, values, regex=False, matchvar='n', valuevar=None):
    """
    Return an expression to match a particular
    field against a list of values, to be used
    within the WHERE clause.

    'values' must be a list, and the generated cypher depends on:
        - the length of the list
        - whether or not it contains 'None'
        - whether or not 'regex' is True
        - whether or not 'valuevar' was given

    If 'valuevar' is given, then the generated cypher will refer to the variable
    instead of the literal values, BUT if the literal values contain None,
    then an additional 'exists()' condition will be added.

    Examples:

        .. code-block: ipython

            In [1]: from neuprint.neuroncriteria import where_expr

            In [2]: where_expr('status', [])
            Out[2]: ''

            In [3]: where_expr('status', [None])
            Out[3]: 'NOT exists(n.status)'

            In [4]: where_expr('status', ['Orphan'])
            Out[4]: "n.status = 'Orphan'"

            In [5]: where_expr('status', ['Orphan', 'Assign'])
            Out[5]: "n.status in ['Orphan', 'Assign']"

            In [6]: where_expr('status', ['Orphan', 'Assign', None])
            Out[6]: "n.status in ['Orphan', 'Assign'] OR NOT exists(n.status)"

            In [7]: where_expr('status', ['Orph.*'], regex=True)
            Out[7]: "n.status =~ 'Orph.*'"

            In [8]: where_expr('instance', ['foo.*', 'bar.*', 'baz.*'], regex=True)
            Out[8]: "n.instance =~ '(foo.*)|(bar.*)|(baz.*)'"

            In [9]: where_expr('bodyId', [123])
            Out[9]: 'n.bodyId = 123'

            In [10]: where_expr('bodyId', [123, 456])
            Out[10]: 'n.bodyId in [123, 456]'

            In [11]: where_expr('bodyId', [123, 456, 789], valuevar='bodies')
            Out[11]: 'n.bodyId in bodies'

            In [12]: where_expr('bodyId', [123, None, 456], valuevar='bodies')
            Out[12]: 'n.bodyId in bodies OR NOT exists(n.bodyId)'

            In [13]: where_expr('status', [IsNull])
            Out[13]: 'n.status IS NULL'

            In [14]: where_expr('status', [NotNull])
            Out[14]: 'n.status NOT NULL'
    """
    assert isinstance(values, collections.abc.Iterable), \
        f"Please pass a list or a variable name, not {values}"

    assert valuevar is None or isinstance(valuevar, str)
    assert not regex or not valuevar, "valuevar is not allowed if using a regex"

    if len(values) == 0:
        return ""

    if len(values) == 1:
        if values[0] is None or values[0] == IsNull:
            return f"NOT exists({matchvar}.{field})"

        if values[0] == NotNull:
            return f"exists({matchvar}.{field})"

        if regex:
            return f"{matchvar}.{field} =~ '{values[0]}'"

        if isinstance(values[0], str):
            return f"{matchvar}.{field} = '{values[0]}'"

        return f"{matchvar}.{field} = {values[0]}"

    if NotNull in values and len(values) > 1:
        raise ValueError('`NotNull` can not be combined with other criteria '
                         'for the same field.')

    # list of values
    if None not in values and IsNull not in values:
        if valuevar:
            return f"{matchvar}.{field} in {valuevar}"
        elif regex:
            assert all(isinstance(v, str) for v in values), \
                "Expected all regex values to be strings"
            r = '|'.join(f'({v})' for v in values)
            return f"{matchvar}.{field} =~ '{r}'"
        else:
            return f"{matchvar}.{field} in {[*values]}"

    # ['some_val', None, 'some_other']
    values = [*filter(lambda v: v not in (None, IsNull), values)]
    if len(values) == 1:
        if regex:
            assert isinstance(values[0], str), \
                "Expected all regex values to be strings"
            return f"{matchvar}.{field} =~ '{values[0]}' OR NOT exists({matchvar}.{field})"
        elif isinstance(values[0], str):
            return f"{matchvar}.{field} = '{values[0]}' OR NOT exists({matchvar}.{field})"
        else:
            return f"{matchvar}.{field} = {values[0]} OR NOT exists({matchvar}.{field})"
    else:
        if regex:
            # Combine the list fo regexes into a single regex
            # of the form: '(regex1)|(regex2)|(regex3)'
            assert all(isinstance(v, str) for v in values), \
                "Expected all regex values to be strings"
            r = '|'.join(f'({v})' for v in values)
            return f"{matchvar}.{field} =~ '{r}' OR NOT exists({matchvar}.{field})"
        elif valuevar:
            return f"{matchvar}.{field} in {valuevar} OR NOT exists({matchvar}.{field})"
        else:
            return f"{matchvar}.{field} in {[*values]} OR NOT exists({matchvar}.{field})"



"""
Miscellaneous plotting functions.


Note:
    These functions require additional dependencies,
    which aren't listed by default dependencies of neuprint-python.
    (See docs for each function.)
"""
import numpy as np
import pandas as pd

from .client import inject_client
from .skeleton import skeleton_df_to_nx

def plot_soma_projections(neurons_df, color_by='cellBodyFiber'):
    """
    Plot the soma locations as XY, XZ, and ZY 2D projections,
    colored by the given column.

    Requires ``bokeh``.

    Returns a layout which can be displayed
    with ``bokeh.plotting.show()``.

    Example:

    .. code-block: python

        from neuprint import fetch_neurons, NeuronCriteria as NC
        from bokeh.plotting import output_notebook
        output_notebook()

        criteria = NC(status='Traced', cropped=False)
        neurons_df, _roi_counts_df = fetch_neurons(criteria)
        p = plot_soma_projections(neurons_df, 'cellBodyFiber')
        show(p)

    """
    import bokeh
    from bokeh.plotting import figure
    from bokeh.layouts import gridplot

    neurons_df = neurons_df[['somaLocation', color_by]].copy()

    extract_soma_coords(neurons_df)
    assign_colors(neurons_df, color_by)

    neurons_with_soma_df = neurons_df.query('not somaLocation.isnull()')
    def soma_projection(axis1, axis2, flip1, flip2):
        x = neurons_with_soma_df[f'soma_{axis1}'].values
        y = neurons_with_soma_df[f'soma_{axis2}'].values
        p = figure(title=f'{axis1}{axis2}')
        p.scatter(x, y, color=neurons_with_soma_df['color'])
        p.x_range.flipped = flip1
        p.y_range.flipped = flip2
        p.toolbar.logo = None
        return p

    p_xy = soma_projection('x', 'y', False, True)
    p_xz = soma_projection('x', 'z', False, True)
    p_zy = soma_projection('z', 'y', True, True)

    # This will produce one big plot with a shared toolbar
    layout = gridplot([[p_xy, p_xz], [None, p_zy]])

    # Discard the help buttons and bokeh logo
    tbar = layout.children[0].toolbar
    tbar.logo = None
    tbar.tools = [t for t in tbar.tools if not isinstance(t, bokeh.models.tools.HelpTool)]

    return layout


def plot_soma_3d(neurons_df, color_by='cellBodyFiber', point_size=1.0):
    """
    Plot the soma locations in 3D, colored randomly according
    to the column given in ``color_by``.

    Requires ``ipyvolume``.
    If using Jupyterlab, install it like this:

    .. code-block: bash

        conda install -c conda-forge ipyvolume
        jupyter labextension install ipyvolume

    Example:

        .. code-block: python

            from neuprint import fetch_neurons, NeuronCriteria as NC

            criteria = NC(status='Traced', cropped=False)
            neurons_df, _roi_counts_df = fetch_neurons(criteria)
            plot_soma_3d(neurons_df, 'cellBodyFiber')
    """
    import ipyvolume.pylab as ipv
    neurons_df = neurons_df[['somaLocation', color_by]].copy()

    extract_soma_coords(neurons_df)
    assign_colors(neurons_df, color_by)

    neurons_with_soma_df = neurons_df.query('not somaLocation.isnull()')
    assert neurons_with_soma_df.eval('color.isnull()').sum() == 0

    soma_x = neurons_with_soma_df['soma_x'].values
    soma_y = neurons_with_soma_df['soma_y'].values
    soma_z = neurons_with_soma_df['soma_z'].values

    def color_to_vals(color_string):
        # Convert bokeh color string into float tuples,
        # e.g. '#00ff00' -> (0.0, 1.0, 0.0)
        s = color_string
        return (int(s[1:3], 16) / 255,
                int(s[3:5], 16) / 255,
                int(s[5:7], 16) / 255 )

    color_vals = neurons_with_soma_df['color'].apply(color_to_vals).tolist()

    # DVID coordinate system assumes (0,0,0) is in the upper-left.
    # For consistency with DVID and neuroglancer conventions,
    # we invert the Y and X coordinates.
    ipv.figure()
    ipv.scatter(soma_x, -soma_y, -soma_z, color=color_vals, marker="circle_2d", size=point_size)
    ipv.show()


@inject_client
def plot_skeleton_3d(skeleton, color='blue', *, client=None):
    """
    Plot the given skeleton in 3D.

    Args:
        skeleton:
            Either a bodyId or a pre-fetched pandas DataFrame

        color:
            See ``ipyvolume`` docs.
            Examples: ``'blue'``, ``'#0000ff'``
            If the skeleton is fragmented, you can give a list
            of colors and each fragment will be shown in a
            different color.

    Requires ``ipyvolume``.
    If using Jupyterlab, install it like this:

    .. code-block: bash

        conda install -c conda-forge ipyvolume
        jupyter labextension install ipyvolume
    """
    import ipyvolume.pylab as ipv

    if np.issubdtype(type(skeleton), np.integer):
        skeleton = client.fetch_skeleton(skeleton, format='pandas')

    assert isinstance(skeleton, pd.DataFrame)
    g = skeleton_df_to_nx(skeleton)

    def skel_path(root):
        """
        We want to plot each skeleton fragment as a single continuous line,
        but that means we have to backtrack: parent -> leaf -> parent
        to avoid jumping from one branch to another.
        This means that the line will be drawn on top of itself,
        and we'll have 2x as many line segments in the plot,
        but that's not a big deal.
        """
        def accumulate_points(n):
            p = (g.nodes[n]['x'], g.nodes[n]['y'], g.nodes[n]['z'])
            points.append(p)

            children = [*g.successors(n)]
            if not children:
                return
            for c in children:
                accumulate_points(c)
                points.append(p)

        points = []
        accumulate_points(root)
        return np.asarray(points)

    # Skeleton may contain multiple fragments,
    # so compute the path for each one.
    def skel_paths(df):
        paths = []
        for root in df.query('link == -1')['rowId']:
            paths.append(skel_path(root))
        return paths

    paths = skel_paths(skeleton)
    if isinstance(color, str):
        colors = len(paths)*[color]
    else:
        colors = (1+len(paths)//len(color))*color

    ipv.figure()
    for points, color in zip(paths, colors):
        ipv.plot(*points.transpose(), color)
    ipv.show()


def extract_soma_coords(neurons_df):
    """
    Expand the ``somaLocation`` column into three separate
    columns for ``soma_x``, ``soma_y``, and ``soma_z``.

    If ``somaLocation is None``, then the soma coords will be ``NaN``.

    Works in-place.
    """
    neurons_df['soma_x'] = neurons_df['soma_y'] = neurons_df['soma_z'] = np.nan

    somaloc = neurons_df.query('not somaLocation.isnull()')['somaLocation']
    somaloc_array = np.asarray(somaloc.tolist())

    neurons_df.loc[somaloc.index, 'soma_x'] = somaloc_array[:, 0]
    neurons_df.loc[somaloc.index, 'soma_y'] = somaloc_array[:, 1]
    neurons_df.loc[somaloc.index, 'soma_z'] = somaloc_array[:, 2]


def assign_colors(neurons_df, color_by='cellBodyFiber'):
    """
    Use a random colortable to assign a color to each row,
    according to the column given in ``color_by``.

    NaN values are always black.

    Works in-place.
    """
    from bokeh.palettes import Turbo256
    colors = list(Turbo256)
    colors[0] = '#000000'
    color_categories = np.sort(neurons_df[color_by].fillna('').unique())
    assert color_categories[0] == ''

    np.random.seed(0)
    np.random.shuffle(color_categories[1:])
    assert color_categories[0] == ''

    while len(colors) < len(color_categories):
        colors.extend(colors[1:])

    color_mapping = dict(zip(color_categories, colors))
    neurons_df['color'] = neurons_df[color_by].fillna('').map(color_mapping)




import os

NEUPRINT_SERVER = 'neuprint.janelia.org'
DATASET = 'hemibrain:v1.2.1'

try:
    TOKEN = os.environ['NEUPRINT_APPLICATION_CREDENTIALS']
except KeyError:
    raise RuntimeError("These tests assume that NEUPRINT_APPLICATION_CREDENTIALS is defined in your environment!")



"""
Functions for performing SPICE-based electrical simulation of a neuron given
its synapses and skeleton using a simple linear passive model.


Try the `interactive simulation tutorial`_ for a tour of basic simulation options.

.. _interactive simulation tutorial: notebooks/SimulationTutorial.ipynb

.. note::

    The ``simulation`` module depends on additional packages.
    Install them from ``conda-forge``:

    .. code-block:: bash

       conda install -c conda-forge ngspice umap-learn scikit-learn matplotlib
"""
# Author: Stephen Plaza
# Delay modeling and spice parsing adapted from code  by Louis K. Scheffer.

import os
import math
import platform
from tempfile import mkstemp
from subprocess import Popen, PIPE, DEVNULL

import numpy as np
import pandas as pd

from scipy.spatial import cKDTree
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage, cut_tree

from .utils import tqdm, UMAP
from .client import inject_client
from .queries import fetch_synapse_connections

# Axon resistance.
Ra_LOW = 0.4
Ra_MED=1.2
Ra_HIGH=4.0

# Membrane resistance.
Rm_LOW = 0.2
Rm_MED=0.8
Rm_HIGH=3.11


class TimingResult:
    def __init__(self, bodyid, delay_matrix, amplitude_matrix, neuron_io, neuron_conn_info, symmetric=False):
        """
        Timing result constructor.

        Provides methods for parsing timing results.

        Args:

            bodyid (int):
                Segment id for neuron

            delay_matrix (dataframe):
                nxm matric of source to sink (if source set == sink set,
                the data can be used to cluster the provided io into different domains.

            neuron_io (dataframe):
                synapse information: location, ispre, brain region
        """
        self.bodyid = bodyid
        self.delay_matrix = delay_matrix
        self.amplitude_matrix = amplitude_matrix
        self.neuron_io = neuron_io
        self.symmetric = symmetric
        self.neuron_conn_info = neuron_conn_info

        # sources = set(delay_matrix.index.to_list())
        # sinks = set(delay_matrix.columns.values())
        # if sources == sinks:
        #     self.symmetric = True

    def compute_region_delay_matrix(self):
        """
        Generate delay and amplitude matrix based on primary brain regions.

        Averages the delay and amplitude between the sources and sinks of the
        timing results.

        Returns:
            (dataframe, dataframe) for the delay and amplitude from brain region to brain region.
        """
        assert(not self.symmetric)

        # determine row and column names
        inrois = set(self.neuron_io[self.neuron_io["io"] == "in"]["roi"].to_list())
        outrois = set(self.neuron_io[self.neuron_io["io"] == "out"]["roi"].to_list())

        inrois = list(inrois)
        outrois = list(outrois)

        inrois.sort()
        outrois.sort()

        delay_matrix = np.zeros((len(inrois), len(outrois)))
        amp_matrix = np.zeros((len(inrois), len(outrois)))

        roi_count = {}
        roi_delays = {}
        roi_amps = {}

        roi2index_in = {}
        roi2index_out = {}
        for idx, roi in enumerate(inrois):
            roi2index_in[roi] = idx
        for idx, roi in enumerate(outrois):
            roi2index_out[roi] = idx

        # roi info
        for drive, row in self.delay_matrix.iterrows():
            inroi = self.neuron_io[self.neuron_io["swcid"] == drive].iloc[0]["roi"]
            for out, val in row.items():
                outroi = self.neuron_io[self.neuron_io["swcid"] == out].iloc[0]["roi"]
                if (inroi, outroi) not in roi_delays:
                    roi_delays[(inroi, outroi)] = 0
                    roi_count[(inroi, outroi)] = 0
                roi_delays[(inroi, outroi)] += val
                roi_count[(inroi, outroi)] += 1

        # calculate average
        for drive, row in self.amplitude_matrix.iterrows():
            inroi = self.neuron_io[self.neuron_io["swcid"] == drive].iloc[0]["roi"]
            for out, val in row.items():
                outroi = self.neuron_io[self.neuron_io["swcid"] == out].iloc[0]["roi"]
                if (inroi, outroi) not in roi_amps:
                    roi_amps[(inroi, outroi)] = 0
                roi_amps[(inroi, outroi)] += val

        for key, val in roi_count.items():
            roi_in, roi_out = key
            idx1 = roi2index_in[roi_in]
            idx2 = roi2index_out[roi_out]
            delay_matrix[idx1, idx2] = roi_delays[key] / val
            amp_matrix[idx1, idx2] = roi_amps[key] / val

        # return dataframes
        return pd.DataFrame(delay_matrix, index=inrois, columns=outrois), pd.DataFrame(amp_matrix, index=inrois, columns=outrois)


    def plot_response_from_region(self, brain_region, path=None):
        """
        Show neuron response to brain regions based on inputs from a given region.

        Args:

            brain_region (str):
                Source brain region.
            path (str):
                Optionally save plot to designated file
        """
        assert(not self.symmetric)

        # matplot for amplitude and delay

        # filter sample brain region inputs
        inputs = self.neuron_io[(self.neuron_io["roi"] == brain_region) & (self.neuron_io["io"] == "in")]["swcid"].to_list()
        delay_matrix_sub = self.delay_matrix[self.delay_matrix.index.isin(inputs)]
        delay_matrix_sub = self.delay_matrix[self.delay_matrix.index.isin(inputs)]
        outrois = set(self.neuron_io[self.neuron_io["io"] == "out"]["roi"].to_list())

        delay_amp_region = []

        for drive, row in delay_matrix_sub.iterrows():
            row2 = self.amplitude_matrix.loc[drive]
            for out, val in row.items():
                outroi = self.neuron_io[self.neuron_io["swcid"] == out].iloc[0]["roi"]
                amp = row2[out]
                delay_amp_region.append([val, amp, outroi])

        plot_df = pd.DataFrame(delay_amp_region, columns=["delay", "amp", "region"])

        # create plot
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        fig.set_figwidth(8)
        fig.set_figheight(8)
        for region in plot_df["region"].unique():
            tdata = plot_df[plot_df["region"] == region]
            ax.scatter(tdata["delay"].to_list(), tdata["amp"].to_list(), c=[np.random.rand(3,)], label=region)
        ax.legend()
        plt.xlabel("delay (ms)")
        plt.ylabel("amplitude (mV)")

        if path is not None:
            plt.savefig(path)

        plt.close()
        return fig


    def plot_neuron_domains(self, path=None):
        """
        Show how the different simulation points cluster in the neuron
        and their corresponding ROI.

        Plots the distance matrix in 2D using delay as the distance.

        Args:
            path (str):
                save plot to file as png
        """

        assert(self.symmetric)

        # ensure matrix is actually symmetric
        delays = self.delay_matrix.values
        for iter1 in range(len(self.delay_matrix)):
            for iter2 in range(iter1, len(self.delay_matrix)):
                if iter1 == iter2:
                    delays[iter1, iter2] = 0
                else:
                    val = (delays[iter1, iter2] + delays[iter2, iter1]) / 2
                    if val < 0:
                        val = 0
                    delays[iter1, iter2] = val
                    delays[iter2, iter1] = val

        # use umap to plot distance matrix
        u = UMAP(metric="precomputed", n_neighbors=90).fit(delays)
        points_2d = u.fit_transform(delays)

        # aasociate a region with each point
        x_y_region = []
        for idx, sid in enumerate(self.delay_matrix.index.to_list()):
            x_y_region.append([points_2d[idx][0], points_2d[idx][1], self.neuron_io[self.neuron_io["swcid"] == sid].iloc[0]["roi"]])

        plot_df = pd.DataFrame(x_y_region, columns=["x", "y", "region"])

        # create plot
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        fig.set_figwidth(8)
        fig.set_figheight(8)
        for region in plot_df["region"].unique():
            tdata = plot_df[plot_df["region"] == region]
            ax.scatter(tdata["x"].to_list(), tdata["y"].to_list(), c=[np.random.rand(3,)], label=region)
        ax.legend()

        if path is not None:
            plt.savefig(path)

        plt.close()
        return fig


    def estimate_neuron_domains(self, num_components, plot=False):
        """
        Estimate the domains based on timing estimates.

        Note: only works for symmetric sink and source simulation.

        Args:
            num_components (int):
                number of cluster domains
            plot (bool):
                If True, create and return a plot that shows cluster results with predicted labels
                (can compare with roi labels from plot_neuron_domains).
                If ``plot`` is a string, it is interpreted as a filepath,
                to which the plot is also written to disk as a PNG.

        Returns:
            (dataframe, dataframe, plot) input and output connection summary split by domain,
            synapse-level neuron_io indicating component partition.
            If a ``plot`` was requested, then the generated plot is returned as the
            third tuple element, otherwise that element is ``None``.
        """

        assert(self.symmetric)

        # ensure matrix is actually symmetric
        delays = self.delay_matrix.values
        for iter1 in range(len(self.delay_matrix)):
            for iter2 in range(iter1, len(self.delay_matrix)):
                if iter1 == iter2:
                    delays[iter1, iter2] = 0
                else:
                    val = (delays[iter1, iter2] + delays[iter2, iter1]) / 2
                    if val < 0:
                        val = 0
                    delays[iter1, iter2] = val
                    delays[iter2, iter1] = val

        # create decomposition using hierarchical cluster over distance matrix
        Dsq = squareform(delays)
        cluster = linkage(Dsq, 'ward')

        if num_components > 1:
            partitions = cut_tree(cluster, n_clusters=num_components)
            best_labels = partitions[:,0]
        else:
            raise RuntimeError("must request 2 or more partitions")
            from sklearn.metrics import silhouette_score
            # find optimal number of clusters using silhouette function or based
            # on provided input
            best_score = -1
            best_labels = None

            for ncomps in range(2, len(self.delay_matrix)//4):
                partitions = cut_tree(cluster, n_clusters=ncomps)
                labels = partitions[:,0]
                score = silhouette_score(delays, labels, metric="precomputed")
                if score > best_score:
                    best_score = score
                    best_labels = labels

        fig = None
        if plot:
            # use umap to plot distance matrix
            u = UMAP(metric="precomputed", n_neighbors=90).fit(delays)
            points_2d = u.fit_transform(delays)

            # aasociate a region with each point
            x_y_region = []
            for idx, sid in enumerate(self.delay_matrix.index.to_list()):
                x_y_region.append([points_2d[idx][0], points_2d[idx][1], best_labels[idx] ])

            plot_df = pd.DataFrame(x_y_region, columns=["x", "y", "region"])

            # create plot
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()
            fig.set_figwidth(8)
            fig.set_figheight(8)
            for region in plot_df["region"].unique():
                tdata = plot_df[plot_df["region"] == region]
                ax.scatter(tdata["x"].to_list(), tdata["y"].to_list(), c=[np.random.rand(3,)], label=region)
            ax.legend()

            if isinstance(plot, str):
                plt.savefig(plot)

            plt.close()

        # build KD tree and associate synapses with each point after the cluster

        filter_list = []
        for drive, row in self.delay_matrix.iterrows():
            idx = self.neuron_io[self.neuron_io["swcid"] == drive].index[0]
            filter_list.append(idx)

        tree = cKDTree(self.neuron_io.iloc[filter_list]["coords"].to_list())
        neuron_conn_info = self.neuron_conn_info.copy()
        coords = list(zip(neuron_conn_info["x"], neuron_conn_info["y"], neuron_conn_info["z"]))
        neuron_conn_info["domain_id"] = [best_labels[tree.query(x)[1]] for x in coords]

        # build summary input / output table

        # find unique input and output groupings
        temp_df = neuron_conn_info.drop_duplicates(subset=["partner", "type", "domain_id"])
        summary_array = []
        for idx, row in temp_df.iterrows():
            io = "input"
            if row["type"] == "pre":
                io = "output"

            matchgroup = neuron_conn_info[(neuron_conn_info["partner"] == row["partner"]) &
                    (neuron_conn_info["type"] == row["type"]) &
                    (neuron_conn_info["domain_id"] == row["domain_id"])]
            count = len(matchgroup)
            rois = list(matchgroup["roi"].unique())
            summary_array.append([io, row["partner"], count, row["domain_id"], rois])
        connection_summary = pd.DataFrame(summary_array, columns=["io", "partner", "weight", "domain_id", "rois"])

        connection_summary = connection_summary.sort_values(by=["io", "weight"], ascending=[False, False]).reset_index(drop=True)
        return connection_summary, neuron_conn_info, fig

class NeuronModel:

    @inject_client
    def __init__(self, bodyid, Ra=Ra_MED, Rm=Rm_MED, Cm=1e-2, *, client=None):
        """
        Neuron model constructor.

        Create model of a neuron which can be simulated.
        in different ways.

        Args:

            bodyid (int):
                Segment id for neuron.
            Ra (float):
                axon resistance
                Examples: 0.4 (``Ra_LOW``), 1.2 (``Ra_MED``), 4.0 (``Ra_HIGH``)
            Rm (float):
                membrane resistance
                Examples: 0.2 (``Rm_LOW``), 0.8 (``Rm_MED``), 3.11 (``Rm_HIGH``)
            Cm (float):
                membrane capacitance (should not very too much between neurons)
        """
        self.bodyid = bodyid


        with tqdm(total=100) as pbar:
            # retrieve healed skeleton
            pbar.set_description("fetching skeleton")
            self.skeleton_df = client.fetch_skeleton(bodyid, heal=True)
            #print("Fetched skeleton")
            pbar.update(20)

            # extract inputs and outputs
            pbar.set_description("fetching output connections")
            outputs = fetch_synapse_connections(bodyid, client=client)
            pbar.update(35)
            pbar.set_description("fetching input connections")
            inputs = fetch_synapse_connections(None, bodyid, client=client)
            pbar.update(35)
            pbar.set_description("creating spice model")
            #print("Fetched synapse connections")

            # combine into one dataframe

            inputs["type"] = ["post"]*len(inputs)
            inputs = inputs[["type", "x_post", "y_post", "z_post", "roi_post", "bodyId_pre" ]].rename(columns={"x_post": "x", "y_post": "y", "z_post": "z", "roi_post": "roi", "bodyId_pre": "partner"})
            inputs["roi"].replace(np.nan, "none", inplace=True)

            input_pins = inputs[["roi"]].copy()
            input_pins["coords"] = list(zip(inputs["x"], inputs["y"], inputs["z"]))
            input_pins["io"] = ["in"]*len(inputs)

            outputs["type"] = ["pre"]*len(outputs)
            outputs = outputs[["type", "x_pre", "y_pre", "z_pre", "roi_pre", "bodyId_post" ]].rename(columns={"x_pre": "x", "y_pre": "y", "z_pre": "z", "roi_pre": "roi", "bodyId_post": "partner"})
            outputs["roi"].replace(np.nan, "none", inplace=True)

            output_pins = outputs[["roi"]].copy()
            output_pins["coords"] = list(zip(outputs["x"], outputs["y"], outputs["z"]))
            output_pins["io"] = ["out"]*len(outputs)

            self.neuron_conn_info = pd.concat([inputs, outputs]).reset_index(drop=True)
            self.io_pins = pd.concat([input_pins, output_pins]).reset_index(drop=True)

            #if len(self.io_pins) == 0:
            #    raise RuntimeError("neuron must have at least 1 inputs or output")

            self.Ra = Ra
            self.Rm = Rm
            self.Cm = 1e-2 # farads per square meeter

            # associate node with each input and output (this will be subsampled later)
            # build kdtree
            tree = cKDTree(list(zip(self.skeleton_df["x"], self.skeleton_df["y"], self.skeleton_df["z"])))
            # apply
            self.io_pins["swcid"] = self.io_pins["coords"].apply(lambda x: tree.query(x)[1]+1)


            # get voxelSize (8e-9) assume nanometers
            self.resolution = client.fetch_custom("MATCH (m :Meta) RETURN m.voxelSize").iloc[0][0][0] * 1e-9

            def build_spice_model():
                #F is the indefinite integral of the area of a cone going from radius r1 to r2 over length L
                def F(x, r1, r2, L):
                    return 2.0*math.pi*(r1*x + (r2-r1)*x**2/(2*L))

                #G is the indefinite integral of the resistance of a cone going from radius r1 to r2 over length L
                def G(x, r1, r2, L):
                    if (r1 != r2):
                        return -(1.0/math.pi) * (L/(r2-r1) * 1.0/(r1 + (r2-r1)*x/L))
                    else:
                        return (1.0/math.pi) * (1.0/r1**2) * x

                # build model
                cs = [0.0] * len(self.skeleton_df)
                rg = [1e30]* len(self.skeleton_df)
                rs = [[0, 0, 1.0] for i in range(len(self.skeleton_df))]   # N-1 Rs, first has none.  node, node, value

                for idx, fromrow in  self.skeleton_df.iterrows():
                    if idx == 0:
                        continue

                    # only one root, should be first entry
                    assert(fromrow["link"] != -1)

                    # row number = link - 1
                    parent = int(fromrow["link"]-1)
                    torow = self.skeleton_df.iloc[parent]

                    # compute axonal resistance
                    L = math.sqrt((fromrow["x"] - torow["x"])**2 + (fromrow["y"] - torow["y"])**2 + (fromrow["z"] - torow["z"])**2) * self.resolution

                    if L == 0:
                        print("L=0 - should not happen")
                        L = 1.0e-9   # set to 1 nm

                    r1 = fromrow["radius"] * self.resolution
                    r2 = torow["radius"] * self.resolution

                    # axonal resistance
                    res = (G(L, r1, r2, L) - G(0, r1, r2, L)) * self.Ra

                    # compute membrane resistance both to and from
                    area_from = F(L/2, r1, r2, L) - F(  0, r1, r2, L)   # Half of segment
                    c_from = area_from * self.Cm
                    rg_from = self.Rm / area_from
                    area_to   = F(L,   r1, r2, L) - F(L/2, r1, r2, L)   # other half
                    c_to   = area_to * self.Cm
                    rg_to = self.Rm / area_to
                    cs[idx] += c_from
                    rg[idx] = rg[idx] * rg_from / (rg[idx] + rg_from)  # in parallel
                    cs[parent] += c_to
                    rg[parent] = rg[parent] * rg_to / (rg[parent] + rg_to)  # in parallel
                    rs[idx][0] = idx
                    rs[idx][1] = parent
                    rs[idx][2] = res
                for i in range(len(cs)):
                    cs[i] = cs[i] * 1000.0  # Convert millisec responses to seconds, to make moments numerically better

                # write-out model string
                modelstr = ""
                for i in range(len(cs)):
                    modelstr += f"C{i+1} {i+1} 0 {cs[i]}\n" # grounded C
                    modelstr += f"RG{i+1} {i+1} 0 {rg[i]}\n" # grounded R membrane resistance
                    assert(rg[i] > 0)
                    if i > 0:
                        modelstr += f"R{i+1} {rs[i][0]+1} {rs[i][1]+1} {rs[i][2]}\n" # axonal resistance
                        assert(rs[i][2] > 0)

                return modelstr

            self.spice_model = build_spice_model()
            pbar.update(10)
            pbar.set_description("built model")
            #print("Built model")

    def _runspice(self, drive, unique_outs):
        """
        Run spice injecting current for a given input and return response for all outputs.

        Note: drive and unique_outs should be swc node ids for an input and output.

        Args:

            drive (int):
                id for input
            unique_outs (list):
                ids for outputs

        Returns:

            Dataframe (output ids, delay, amplitude)
        """

        # apply current at the specified input location
        drive_str = f"RDRIVE {drive} {len(self.skeleton_df)+1} 10000000000\n" # 0.1 ns conductance
        drive_str += f"V1 {len(self.skeleton_df)+1} 0 EXP(0 60.0 0.1 0.1 1.1 1.0 40)\n"
        drive_str += ".tran 0.1 40\n" # work from 0-10 ms (try 40)
        drive_str += ".options filetype=binary\n"

        # call command line spice simulator and write to temporary file
        fd, path = mkstemp()

        if platform.system() == "Windows":
            ngspice = "ngspice_con.exe"
        else:
            ngspice = "ngspice"

        # run ngspice
        try:
            p = Popen([ngspice, "-b", "-r", path], stdin=PIPE, stdout=DEVNULL, stderr=DEVNULL)
        except FileNotFoundError as ex:
            msg = ("The 'ngspice' circuit simulation tool is not installed (or not on your PATH).\n\n"
                   "Please install it:\n\n"
                   "  conda install -c conda-forge ngspice\n\n")
            raise RuntimeError(msg) from ex

        title = 'FIBSEM simulation\n'
        data = title + self.spice_model + drive_str
        p.communicate(data.encode())

        """Read ngspice binary raw files. Return tuple of the data, and the
        plot metadata. The dtype of the data contains field names. This is
        not very robust yet, and only supports ngspice.

        # Example header of raw file
        # Title: rc band pass example circuit
        # Date: Sun Feb 21 11:29:14  2016
        # Plotname: AC Analysis
        # Flags: complex
        # No. Variables: 3
        # No. Points: 41
        # Variables:
        #         0       frequency       frequency       grid=3
        #         1       v(out)  voltage
        #         2       v(in)   voltage
        # Binary:
        """
        sim_results = None
        BSIZE_SP = 512 # Max size of a line of data; we don't want to read the
                        # whole file to find a line, in case file does not have
                        # expected structure.
        MDATA_LIST = [b'title', b'date', b'plotname', b'flags', b'no. variables', b'no. points', b'dimensions', b'command', b'option']
        with os.fdopen(fd, 'rb') as fp:
            plot = {}
            count = 0
            arrs = []
            plots = []
            while (True):
                try:
                    mdata = fp.readline(BSIZE_SP).split(b':', maxsplit=1)
                except:
                    raise RuntimeError("cannot parse spice output")
                if len(mdata) == 2:
                    if mdata[0].lower() in MDATA_LIST:
                        plot[mdata[0].lower()] = mdata[1].strip()
                    if mdata[0].lower() == b'variables':
                        nvars = int(plot[b'no. variables'])
                        npoints = int(plot[b'no. points'])
                        plot['varnames'] = []
                        plot['varunits'] = []
                        for varn in range(nvars):
                            varspec = (fp.readline(BSIZE_SP).strip()
                                       .decode('ascii').split())
                            assert(varn == int(varspec[0]))
                            plot['varnames'].append(varspec[1])
                            plot['varunits'].append(varspec[2])
                    if mdata[0].lower() == b'binary':
                        rowdtype = np.dtype({'names': plot['varnames'],
                                             'formats': [np.complex_ if b'complex'
                                                         in plot[b'flags']
                                                         else np.float_]*nvars})
                        # We should have all the metadata by now
                        arrs.append(np.fromfile(fp, dtype=rowdtype, count=npoints))
                        plots.append(plot)
                        fp.readline() # Read to the end of line
                else:
                    break

            # only one analysis
            sim_results = arrs[0]
        # delete file
        os.unlink(path)

        times = [0.0] * len(sim_results)
        # for each output (based on skeleton compartment id),
        # determine the delay to the max voltage and the width
        out_results = []

        for i in range(len(sim_results)):
            times[i] = sim_results[i]["time"]
            # parse out delay and amplitdue responses
        volts = [0.0] * len(sim_results)

        for idx in range(len(unique_outs)):
            # compartment id
            j = unique_outs[idx]

            for i in range(len(sim_results)):
                volts[i] = sim_results[i][j]

            #Now find max voltage, time at max voltag,
            maxv = 0.0
            maxt = -1
            for i in range(1,len(volts)-1):
                if volts[i-1] < volts[i] and volts[i] > volts[i+1]:
                    #interpolate to find max
                    v1 = volts[i-1] - volts[i]  #relative value at t1 (negative)
                    v2 = volts[i+1] - volts[i]  #same for t2
                    x1 = times[i-1] - times[i]
                    x2 = times[i+1] - times[i]
                    a = (v1 - v2*x1/x2)/(x1**2 - x1*x2)
                    b = (v1 - a*x1**2)/x1
                    deltax = -b/(2*a)
                    maxv = volts[i]
                    maxt = times[i] + deltax
                    break     # should be only one peak.
            assert(maxt >= 0)

            maxt -= 1.1   # subtract the peak time of the input
            for i in range(len(volts)):
                if volts[i] > maxv/2:
                    break
            for k in reversed(range(len(volts))):
                if volts[k] > maxv/2:
                    break
            out_results.append([j, maxt, maxv, times[k] - times[i]])
            assert(maxt <= 20)

        return pd.DataFrame(out_results, columns=["comp id", "delay", "maxv", "width"])


    def simulate(self, max_per_region=10):
        """Simulate passive model based on neuron inputs and outputs.

        Args:

            max_per_region (int):
                maxinum number of inputs per primary ROI to sample (0 means all)

        Returns:

            TimingResult (contains input/output delay matrix)
        """


        # only grab unique skel comp id rows and randomize data
        unique_io = self.io_pins.drop_duplicates(subset=["swcid"]).sample(frac=1).reset_index(drop=True)

        # grab top max_per region per input and all outputs
        drive_list = []
        unique_outs = []

        # number of drives per input
        rcounter = {}
        for idx, row in unique_io.iterrows():
            if row["io"] == "out":
                unique_outs.append(row["swcid"])
            else:
                if row["roi"] not in rcounter:
                    rcounter[row["roi"]] = 0
                if max_per_region > 0 and rcounter[row["roi"]] == max_per_region:
                    continue
                rcounter[row["roi"]] += 1
                drive_list.append(row["swcid"])
        if len(unique_outs) == 0:
            raise RuntimeError("neuron must have at least one output")
        if len(drive_list) == 0:
            raise RuntimeError("neuron must specify at least one input")

        # simulate each drive (provide progress bar)
        delay_data = []
        amp_data = []
        for drive in tqdm(drive_list):
            # run simulation
            sim_results = self._runspice(drive, unique_outs)

            delay_data.append(sim_results["delay"].to_list())
            amp_data.append(sim_results["maxv"].to_list())

        delay_df = pd.DataFrame(delay_data, columns=unique_outs, index=drive_list)
        amp_df = pd.DataFrame(amp_data, columns=unique_outs, index=drive_list)

        # return simulation results
        return TimingResult(self.bodyid, delay_df, amp_df, self.io_pins, self.neuron_conn_info, False)


    def estimate_intra_neuron_delay(self, num_points=100):
        """Simulate delay between random parts of a neuron.

        This function produces a delay matrix for a set of points
        determined by pre or post-synaptic sites.  The result
        is a square distance matrix which can be clustered to
        determine isopotential regions.  The points where current
        are injected do not represent real neural mechanisms
        and in practice random points could be chosen for simulation.
        Synapse location are chosen for convenience and to ensure
        proper weighting on 'important' parts of the neuron.

        Args:

            num_points(int):
                number of points to simulate.

        Returns:

            TimingResult (contains input/output delay matrix)
        """
        if num_points < 10:
            raise RuntimeError("must specifiy at least 10 simulation points")

        # only grab unique skel comp id rows and randomize data
        unique_io = self.io_pins.drop_duplicates(subset=["swcid"]).sample(frac=1).reset_index(drop=True)

        # consider input/output response symmetrically to create a distance
        io_list = unique_io[unique_io["io"] == "out"][0:(num_points//2)]["swcid"].to_list()
        io_list.extend(unique_io[unique_io["io"] == "in"][0:(num_points//2)]["swcid"].to_list())

        # simulate each drive (provide progress bar)
        delay_data = []
        amp_data = []
        for drive in tqdm(io_list):
            # run simulation
            sim_results = self._runspice(drive, io_list)

            delay_data.append(sim_results["delay"].to_list())
            amp_data.append(sim_results["maxv"].to_list())

        delay_df = pd.DataFrame(delay_data, columns=io_list, index=io_list)
        amp_df = pd.DataFrame(amp_data, columns=io_list, index=io_list)

        # return simulation results
        return TimingResult(self.bodyid, delay_df, amp_df, self.io_pins, self.neuron_conn_info, True)



import pytest
import numpy as np
import pandas as pd
import networkx as nx

from neuprint import Client, default_client, set_default_client
from neuprint import (fetch_skeleton, heal_skeleton, reorient_skeleton, skeleton_df_to_nx, skeleton_df_to_swc, skeleton_swc_to_df)

from neuprint.tests import NEUPRINT_SERVER, DATASET


@pytest.fixture(scope='module')
def client():
    c = Client(NEUPRINT_SERVER, DATASET)
    set_default_client(c)
    assert default_client() is c
    return c


@pytest.fixture
def linear_skeleton():
    """
    A test fixture to produce a fake 'skeleton'
    with no branches, just 10 nodes in a line.
    """
    rows = np.arange(1,11)
    coords = np.zeros((10,3), dtype=int)
    coords[:,0] = rows**2
    radii = rows.astype(np.float32)
    links = [-1, *range(1,10)]

    df = pd.DataFrame({'rowId': rows,
                       'x': coords[:,0],
                       'y': coords[:,1],
                       'z': coords[:,2],
                       'radius': radii,
                       'link': links})
    return df


def test_skeleton_df_to_nx(linear_skeleton):
    g = skeleton_df_to_nx(linear_skeleton, directed=False)
    assert not isinstance(g, nx.DiGraph)
    expected_edges = linear_skeleton[['rowId', 'link']].values[1:]
    expected_edges.sort(axis=1)
    assert (np.array(g.edges) == expected_edges).all()

    g = skeleton_df_to_nx(linear_skeleton, directed=True)
    assert isinstance(g, nx.DiGraph)
    assert (np.array(g.edges) == linear_skeleton[['rowId', 'link']].values[1:]).all()

    g = skeleton_df_to_nx(linear_skeleton, with_attributes=True)
    assert (np.array(g.edges) == linear_skeleton[['rowId', 'link']].values[1:]).all()
    for row in linear_skeleton.itertuples():
        attrs = g.nodes[row.rowId]
        assert tuple(attrs[k] for k in [*'xyz', 'radius']) == (row.x, row.y, row.z, row.radius)


def test_skeleton_df_to_swc(linear_skeleton):
    swc = skeleton_df_to_swc(linear_skeleton)
    roundtrip_df = skeleton_swc_to_df(swc)
    assert (roundtrip_df == linear_skeleton).all().all()


def test_reorient_skeleton(linear_skeleton):
    s = linear_skeleton.copy()
    reorient_skeleton(s, 10)
    assert (s['link'] == [*range(2,11), -1]).all()

    s = linear_skeleton.copy()
    reorient_skeleton(s, xyz=(100,0,0))
    assert (s['link'] == [*range(2,11), -1]).all()

    s = linear_skeleton.copy()
    reorient_skeleton(s, use_max_radius=True)
    assert (s['link'] == [*range(2,11), -1]).all()


def test_reorient_broken_skeleton(linear_skeleton):
    broken_skeleton = linear_skeleton.copy()
    broken_skeleton.loc[2, 'link'] = -1
    broken_skeleton.loc[7, 'link'] = -1

    s = broken_skeleton.copy()
    reorient_skeleton(s, 10)
    assert (s['link'].iloc[7:10] == [9,10,-1]).all()

    # reorienting shouldn't change the number of roots,
    # though they may change locations.
    assert len(s.query('link == -1')) == 3


def test_heal_skeleton(linear_skeleton):
    broken_skeleton = linear_skeleton.copy()
    broken_skeleton.loc[2, 'link'] = -1
    broken_skeleton.loc[7, 'link'] = -1

    healed_skeleton = heal_skeleton(broken_skeleton)
    assert (healed_skeleton == linear_skeleton).all().all()


def test_heal_skeleton_with_threshold(linear_skeleton):
    broken_skeleton = linear_skeleton.copy()
    broken_skeleton.loc[2, 'link'] = -1
    broken_skeleton.loc[7, 'link'] = -1

    healed_skeleton = heal_skeleton(broken_skeleton, 10.0)

    # With a threshold of 10, the first break could be healed,
    # but not the second.
    expected_skeleton = linear_skeleton.copy()
    expected_skeleton.loc[7, 'link'] = -1
    assert (healed_skeleton == expected_skeleton).all().all()


def test_fetch_skeleton(client):
    orig_df = fetch_skeleton(5813027016, False)
    healed_df = fetch_skeleton(5813027016, True)

    assert len(orig_df) == len(healed_df)
    assert (healed_df['link'] == -1).sum() == 1
    assert healed_df['link'].iloc[0] == -1


@pytest.mark.skip("Need to write a test for skeleton_segments()")
def test_skeleton_segments(linear_skeleton):
    pass


if __name__ == "__main__":
    args = ['-s', '--tb=native', '--pyargs', 'neuprint.tests.test_skeleton']
    #args += ['-k', 'heal_skeleton']
    pytest.main(args)



from setuptools import setup, find_packages

import versioneer

with open('dependencies.txt') as f:
    requirements = f.read().splitlines()
    requirements = [l for l in requirements if not l.strip().startswith('#')]

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='neuprint-python',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Python client utilties for interacting with the neuPrint connectome analysis service",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Stuart Berg",
    author_email='bergs@hhmi.janelia.org',
    url='https://github.com/connectome-neuprint/neuprint-python',
    packages=find_packages(),
    entry_points={},
    install_requires=requirements,
    keywords='neuprint-python',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



import pytest
import pandas as pd
from neuprint import Client, default_client, set_default_client
from neuprint.client import inject_client
from neuprint.tests import NEUPRINT_SERVER, DATASET

EXAMPLE_BODY = 5813037876 # Delta6G, Delta6G_04, Traced, non-cropped


def test_members():
    set_default_client(None)
    assert default_client() is None
    c = Client(NEUPRINT_SERVER, DATASET)
    assert c.server == f'https://{NEUPRINT_SERVER}'
    assert c.dataset == DATASET

    assert default_client() is c

    df = c.fetch_custom("MATCH (m:Meta) RETURN m.primaryRois as rois")
    assert isinstance(df, pd.DataFrame)
    assert df.columns == ['rois']
    assert len(df) == 1
    assert isinstance(df['rois'].iloc[0], list)

    assert isinstance(c.fetch_available(), list)
    assert isinstance(c.fetch_help(), str)
    assert c.fetch_server_info() is True
    assert isinstance(c.fetch_version(), str)
    assert isinstance(c.fetch_database(), dict)
    assert isinstance(c.fetch_datasets(), dict)
    assert isinstance(c.fetch_db_version(), str)
    assert isinstance(c.fetch_profile(), dict)
    assert isinstance(c.fetch_token(), str)
    assert isinstance(c.fetch_daily_type(), tuple)
    assert isinstance(c.fetch_roi_completeness(), pd.DataFrame)
    assert isinstance(c.fetch_roi_connectivity(), pd.DataFrame)
    assert isinstance(c.fetch_roi_mesh('AB(R)'), bytes)
    assert isinstance(c.fetch_skeleton(EXAMPLE_BODY), pd.DataFrame)
    assert isinstance(c.fetch_neuron_keys(), list)


def test_fetch_skeleton():
    c = Client(NEUPRINT_SERVER, DATASET)
    orig_df = c.fetch_skeleton(5813027016, False)
    healed_df = c.fetch_skeleton(5813027016, True)

    assert len(orig_df) == len(healed_df)
    assert (healed_df['link'] == -1).sum() == 1
    assert healed_df['link'].iloc[0] == -1


@pytest.mark.xfail
def test_broken_members():
    """
    These endpoints are listed in the neuprintHTTP API,
    but don't seem to work.
    """
    c = Client(NEUPRINT_SERVER, DATASET)

    # Broken. neuprint returns error 500
    assert isinstance(c.fetch_instances(), list)


@pytest.mark.skip
def test_keyvalue():
    # TODO:
    # What is an appropriate key/value to test with?
    c = Client(NEUPRINT_SERVER, DATASET)
    c.post_raw_keyvalue(instance, key, b'test-test-test')
    c.fetch_raw_keyvalue(instance, key)


def test_inject_client():
    c = Client(NEUPRINT_SERVER, DATASET)
    c2 = Client(NEUPRINT_SERVER, DATASET)

    set_default_client(c)

    @inject_client
    def f(*, client):
        return client

    # Uses default client unless client was specified
    assert f() is c
    assert f(client=c2) is c2

    with pytest.raises(AssertionError):
        # Wrong signature -- asserts
        @inject_client
        def f2(client):
            pass

if __name__ == "__main__":
    args = ['-s', '--tb=native', '--pyargs', 'neuprint.tests.test_client']
    #args += ['-k', 'fetch_skeleton']
    pytest.main(args)



from ..client import inject_client


@inject_client
def fetch_custom(cypher, dataset="", format='pandas', *, client=None):
    '''
    Make a custom cypher query.

    Alternative form of :py:meth:`.Client.fetch_custom()`, as a free function.
    That is, ``fetch_custom(..., client=c)`` is equivalent to ``c.fetch_custom(...)``.

    If ``client=None``, the default ``Client`` is used
    (assuming you have created at least one ``Client``.)

    Args:
        cypher:
            A cypher query string

        dataset:
            *Deprecated. Please provide your dataset as a Client constructor argument.*

            Which neuprint dataset to query against.
            If None provided, the client's default dataset is used.

        format:
            Either 'pandas' or 'json'.
            Whether to load the results into a pandas DataFrame,
            or return the server's raw json response as a Python dict.

        client:
            If not provided, the global default :py:class:`.Client` will be used.

    Returns:
        Either json or DataFrame, depending on ``format``.

    .. code-block:: ipython

        In [4]: from neuprint import fetch_custom
           ...:
           ...: q = """\\
           ...: MATCH (n:Neuron)
           ...: WHERE n.bodyId = 5813027016
           ...: RETURN n.type, n.instance
           ...: """
           ...: fetch_custom(q)
        Out[4]:
          n.type      n.instance
        0   FB4Y  FB4Y(EB/NO1)_R
    '''
    return client.fetch_custom(cypher, dataset, format)


@inject_client
def fetch_meta(*, client=None):
    """
    Fetch the dataset metadata.
    Parses json fields as needed.

    Returns:
        dict

    Example

    .. code-block:: ipython

        In [1]: from neuprint import fetch_meta

        In [2]: meta = fetch_meta()

        In [3]: list(meta.keys())
        Out[3]:
        ['dataset',
         'info',
         'lastDatabaseEdit',
         'latestMutationId',
         'logo',
         'meshHost',
         'neuroglancerInfo',
         'neuroglancerMeta',
         'postHPThreshold',
         'postHighAccuracyThreshold',
         'preHPThreshold',
         'primaryRois',
         'roiHierarchy',
         'roiInfo',
         'statusDefinitions',
         'superLevelRois',
         'tag',
         'totalPostCount',
         'totalPreCount',
         'uuid']
    """
    q = """\
        MATCH (m:Meta)
        WITH m as m,
             apoc.convert.fromJsonMap(m.roiInfo) as roiInfo,
             apoc.convert.fromJsonMap(m.roiHierarchy) as roiHierarchy,
             apoc.convert.fromJsonMap(m.neuroglancerInfo) as neuroglancerInfo,
             apoc.convert.fromJsonList(m.neuroglancerMeta) as neuroglancerMeta,
             apoc.convert.fromJsonMap(m.statusDefinitions) as statusDefinitions
        RETURN m as meta, roiInfo, roiHierarchy, neuroglancerInfo, neuroglancerMeta, statusDefinitions
    """
    df = client.fetch_custom(q)
    meta = df['meta'].iloc[0]
    meta.update(df.drop(columns='meta').iloc[0].to_dict())
    return meta



import sys
import copy
from textwrap import dedent

import numpy as np
import pandas as pd

from ..client import inject_client
from ..utils import tqdm, iter_batches
from .neuroncriteria import NeuronCriteria, neuroncriteria_args
from .synapsecriteria import SynapseCriteria
from .connectivity import fetch_adjacencies


@inject_client
@neuroncriteria_args('neuron_criteria')
def fetch_synapses(neuron_criteria, synapse_criteria=None, batch_size=10, *, client=None):
    """
    Fetch synapses from a neuron or selection of neurons.

    Args:

        neuron_criteria (bodyId(s), type/instance, or :py:class:`.NeuronCriteria`):
            Determines which bodies to fetch synapses for.

            Note:
                Any ROI criteria specified in this argument does not affect
                which synapses are returned, only which bodies are inspected.

        synapse_criteria (SynapseCriteria):
            Optional. Allows you to filter synapses by roi, type, confidence.
            See :py:class:`.SynapseCriteria` for details.

            If the criteria specifies ``primary_only=True`` only primary ROIs will be returned in the results.
            If a synapse does not intersect any primary ROI, it will be listed with an roi of ``None``.
            (Since 'primary' ROIs do not overlap, each synapse will be listed only once.)
            Otherwise, all ROI names will be included in the results.
            In that case, some synapses will be listed multiple times -- once per intersecting ROI.
            If a synapse does not intersect any ROI, it will be listed with an roi of ``None``.

        batch_size:
            To improve performance and avoid timeouts, the synapses for multiple bodies
            will be fetched in batches, where each batch corresponds to N bodies.
            This argument sets the batch size N.

        client:
            If not provided, the global default :py:class:`.Client` will be used.

    Returns:

        DataFrame in which each row represent a single synapse.
        Unless ``primary_only`` was specified, some synapses may be listed more than once,
        if they reside in more than one overlapping ROI.

    Example:

        .. code-block:: ipython

            In [1]: from neuprint import NeuronCriteria as NC, SynapseCriteria as SC, fetch_synapses
               ...: fetch_synapses(NC(type='ADL.*', rois=['FB']),
               ...:                SC(rois=['LH(R)', 'SIP(R)'], primary_only=True))
            Out[1]:
                    bodyId  type     roi      x      y      z  confidence
            0   5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            1   5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            2   5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            3   5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            4   5812983094   pre   LH(R)   5447  21281  19155    0.991000
            5   5812983094   pre   LH(R)   5434  21270  19201    0.995000
            6   5812983094   pre   LH(R)   5434  21270  19201    0.995000
            7   5812983094   pre   LH(R)   5434  21270  19201    0.995000
            8   5812983094   pre   LH(R)   5447  21281  19155    0.991000
            9   5812983094   pre   LH(R)   5447  21281  19155    0.991000
            10  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            11  5812983094   pre   LH(R)   5434  21270  19201    0.995000
            12  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            13  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            14  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            15  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            16  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            17  5812983094   pre   LH(R)   5434  21270  19201    0.995000
            18  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            19  5812983094   pre   LH(R)   5434  21270  19201    0.995000
            20  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            21  5812983094   pre   LH(R)   5434  21270  19201    0.995000
            22  5812983094   pre   LH(R)   5447  21281  19155    0.991000
            23  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            24  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            25  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            26  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            27  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            28  5812983094   pre  SIP(R)  15300  25268  14043    0.992000
            29   764377593  post   LH(R)   8043  21587  15146    0.804000
            30   764377593  post   LH(R)   8057  21493  15140    0.906482
            31   859152522   pre  SIP(R)  13275  25499  13629    0.997000
            32   859152522   pre  SIP(R)  13275  25499  13629    0.997000
            33   859152522  post  SIP(R)  13349  25337  13653    0.818386
            34   859152522  post  SIP(R)  12793  26362  14202    0.926918
            35   859152522   pre  SIP(R)  13275  25499  13629    0.997000

    """
    neuron_criteria.matchvar = 'n'
    q = f"""
        {neuron_criteria.global_with(prefix=8)}
        MATCH (n:{neuron_criteria.label})
        {neuron_criteria.all_conditions(prefix=8)}
        RETURN n.bodyId as bodyId
    """
    bodies = client.fetch_custom(q)['bodyId'].values

    batch_dfs = []
    for batch_bodies in tqdm(iter_batches(bodies, batch_size)):
        batch_criteria = copy.copy(neuron_criteria)
        batch_criteria.bodyId = batch_bodies
        batch_df = _fetch_synapses(batch_criteria, synapse_criteria, client)
        batch_dfs.append( batch_df )

    if batch_dfs:
        return pd.concat( batch_dfs, ignore_index=True )

    # Return empty results, but with correct dtypes
    dtypes = {'bodyId': np.dtype('int64'),
              'type': pd.CategoricalDtype(categories=['pre', 'post'], ordered=False),
              'roi': np.dtype('O'),
              'x': np.dtype('int32'),
              'y': np.dtype('int32'),
              'z': np.dtype('int32'),
              'confidence': np.dtype('float32')}

    return pd.DataFrame([], columns=dtypes.keys()).astype(dtypes)


def _fetch_synapses(neuron_criteria, synapse_criteria, client):
    neuron_criteria.matchvar = 'n'

    if synapse_criteria is None:
        synapse_criteria = SynapseCriteria()

    if synapse_criteria.primary_only:
        return_rois = {*client.primary_rois}
    else:
        return_rois = {*client.all_rois}

    # If the user specified rois to filter synapses by, but hasn't specified rois
    # in the NeuronCriteria, add them to the NeuronCriteria to speed up the query.
    if synapse_criteria.rois and not neuron_criteria.rois:
        neuron_criteria.rois = {*synapse_criteria.rois}
        neuron_criteria.roi_req = 'any'

    # Fetch results
    cypher = dedent(f"""\
        {neuron_criteria.global_with(prefix=8)}
        MATCH (n:{neuron_criteria.label})
        {neuron_criteria.all_conditions('n', prefix=8)}

        MATCH (n)-[:Contains]->(ss:SynapseSet),
              (ss)-[:Contains]->(s:Synapse)

        {synapse_criteria.condition('n', 's', prefix=8)}
        // De-duplicate 's' because 'pre' synapses can appear in more than one SynapseSet
        WITH DISTINCT n, s

        RETURN n.bodyId as bodyId,
               s.type as type,
               s.confidence as confidence,
               s.location.x as x,
               s.location.y as y,
               s.location.z as z,
               apoc.map.removeKeys(s, ['location', 'confidence', 'type']) as syn_info
    """)
    data = client.fetch_custom(cypher, format='json')['data']

    # Assemble DataFrame
    syn_table = []
    for body, syn_type, conf, x, y, z, syn_info in data:
        # Exclude non-primary ROIs if necessary
        syn_rois = return_rois & {*syn_info.keys()}
        # Fixme: Filter for the user's ROIs (drop duplicates)
        for roi in syn_rois:
            syn_table.append((body, syn_type, roi, x, y, z, conf))

        if not syn_rois:
            syn_table.append((body, syn_type, None, x, y, z, conf))

    syn_df = pd.DataFrame(syn_table, columns=['bodyId', 'type', 'roi', 'x', 'y', 'z', 'confidence'])

    # Save RAM with smaller dtypes and interned strings
    syn_df['type'] = pd.Categorical(syn_df['type'], ['pre', 'post'])
    syn_df['roi'] = syn_df['roi'].apply(lambda s: sys.intern(s) if s else s)
    syn_df['x'] = syn_df['x'].astype(np.int32)
    syn_df['y'] = syn_df['y'].astype(np.int32)
    syn_df['z'] = syn_df['z'].astype(np.int32)
    syn_df['confidence'] = syn_df['confidence'].astype(np.float32)

    return syn_df


@inject_client
@neuroncriteria_args('neuron_criteria')
def fetch_mean_synapses(neuron_criteria, synapse_criteria=None, batch_size=100, *, client=None):
    """
    Fetch per-ROI average synapse position and confidence for a set of neurons.

    Args:

        neuron_criteria (bodyId(s), type/instance, or :py:class:`.NeuronCriteria`):
            Determines which bodies to fetch synapses for.

            Note:
                Any ROI criteria specified in this argument does not affect
                which synapses are returned, only which bodies are inspected.

        synapse_criteria (SynapseCriteria):
            Optional. Allows you to filter synapses by roi, type, confidence.
            See :py:class:`.SynapseCriteria` for details.

            If the criteria specifies ``primary_only=True`` only primary ROIs will be returned in the results.
            If a synapse does not intersect any primary ROI, it will be ignored by this function.

        batch_size:
            To improve performance and avoid timeouts, the synapses for multiple bodies
            will be processed in batches, where each batch corresponds to N bodies.
            This argument sets the batch size N.

        client:
            If not provided, the global default :py:class:`.Client` will be used.

    Returns:

        DataFrame in which each row contains the average synapse position and average
        confidence for a particular body, ROI, and synapse type (pre/post).

    Example:

        .. code-block:: ipython

            In [1]: from neuprint import fetch_mean_synapses, SynapseCriteria as SC
               ...: fetch_mean_synapses('LC10', SC(type='pre', rois=['LO(R)', 'AOTU(R)']))
            Out[1]:
                    bodyId      roi type  count             x             y             z  confidence
            0    1017448980  AOTU(R)  pre    141  10691.737305  30304.355469  15508.099609    0.956000
            1    1017448980    LO(R)  pre      2   8530.500000  18898.000000  34857.000000    0.954500
            2    1017090133  AOTU(R)  pre     49  10560.469727  31032.041016  14548.550781    0.956286
            3    1017090133    LO(R)  pre     52   3737.115479  19489.923828  29530.000000    0.949673
            4    1017094185  AOTU(R)  pre    157  10922.248047  30326.693359  15349.484375    0.959185
            ..          ...      ...  ...    ...           ...           ...           ...         ...
            772  1262602271    LO(R)  pre      3   5178.666504  14584.666992  23227.333984    0.957333
            773  1291573712  AOTU(R)  pre    273  11285.106445  31097.925781  15987.640625    0.954418
            774  1291573712    LO(R)  pre      3   6482.000000  13845.333008  25962.666016    0.936333
            775  5812993252  AOTU(R)  pre     17  11110.706055  30892.470703  17220.175781    0.960471
            776  5812993252    LO(R)  pre     39   7758.897461  11723.154297  28176.794922    0.943128
    """
    neuron_criteria.matchvar = 'n'
    q = f"""
        {neuron_criteria.global_with(prefix=8)}
        MATCH (n:{neuron_criteria.label})
        {neuron_criteria.all_conditions(prefix=8)}
        RETURN n.bodyId as bodyId
    """
    bodies = client.fetch_custom(q)['bodyId'].values

    batch_dfs = []
    for batch_bodies in tqdm(iter_batches(bodies, batch_size)):
        batch_criteria = copy.copy(neuron_criteria)
        batch_criteria.bodyId = batch_bodies
        batch_df = _fetch_mean_synapses(batch_criteria, synapse_criteria, client)
        batch_dfs.append( batch_df )

    if batch_dfs:
        return pd.concat( batch_dfs, ignore_index=True )

    # Return empty results, but with correct dtypes
    dtypes = {'bodyId': np.dtype('int64'),
              'type': pd.CategoricalDtype(categories=['pre', 'post'], ordered=False),
              'roi': np.dtype('O'),
              'count': np.dtype('int32'),
              'x': np.dtype('float32'),
              'y': np.dtype('float32'),
              'z': np.dtype('float32')}

    return pd.DataFrame([], columns=dtypes.keys()).astype(dtypes)


def _fetch_mean_synapses(neuron_criteria, synapse_criteria, client):
    neuron_criteria.matchvar = 'n'

    if synapse_criteria is None:
        synapse_criteria = SynapseCriteria()

    if synapse_criteria.rois:
        rois = synapse_criteria.rois
    elif synapse_criteria.primary_only:
        rois = client.primary_rois
    else:
        rois = client.all_rois

    # If the user specified rois to filter synapses by, but hasn't specified rois
    # in the NeuronCriteria, add them to the NeuronCriteria to speed up the query.
    if synapse_criteria.rois and not neuron_criteria.rois:
        neuron_criteria.rois = {*synapse_criteria.rois}
        neuron_criteria.roi_req = 'any'

    # Fetch results
    cypher = dedent(f"""\
        WITH {rois} as rois
        {neuron_criteria.global_with('rois', prefix=8)}
        MATCH (n:{neuron_criteria.label})
        {neuron_criteria.all_conditions('n', prefix=8)}

        MATCH (n)-[:Contains]->(ss:SynapseSet),
              (ss)-[:Contains]->(s:Synapse)

        {synapse_criteria.condition('rois', 'n', 's', prefix=8)}
        // De-duplicate 's' because 'pre' synapses can appear in more than one SynapseSet
        WITH DISTINCT rois, n, s

        // Extract properties as rows
        UNWIND KEYS(s) as roi

        // Filter for the ROIs we need
        WITH rois, n, s, roi
        WHERE roi in {rois}

        RETURN n.bodyId as bodyId,
               roi,
               s.type as type,
               count(s) as count,
               avg(s.location.x) as x,
               avg(s.location.y) as y,
               avg(s.location.z) as z,
               avg(s.confidence) as confidence
    """)
    syn_df = client.fetch_custom(cypher)

    # Save RAM with smaller dtypes and interned strings
    syn_df['type'] = pd.Categorical(syn_df['type'], ['pre', 'post'])
    syn_df['roi'] = syn_df['roi'].apply(lambda s: sys.intern(s) if s else s)
    syn_df['count'] = syn_df['count'].astype(np.int32)
    syn_df['x'] = syn_df['x'].astype(np.float32)
    syn_df['y'] = syn_df['y'].astype(np.float32)
    syn_df['z'] = syn_df['z'].astype(np.float32)
    syn_df['confidence'] = syn_df['confidence'].astype(np.float32)

    return syn_df


@inject_client
@neuroncriteria_args('source_criteria', 'target_criteria')
def fetch_synapse_connections(source_criteria=None, target_criteria=None, synapse_criteria=None, min_total_weight=1, batch_size=10_000, *, client=None):
    """
    Fetch synaptic-level connections between source and target neurons.

    Note:
        Use this function if you need information about individual synapse connections,
        such as their exact positions or confidence scores.
        If you're just interested in aggregate neuron-to-neuron connection info
        (including connection strengths and ROI intersections), see
        :py:func:`fetch_simple_connections()` and :py:func:`fetch_adjacencies()`,
        which are faster and have more condensed outputs than this function.

    Note:
        If you experience timeouts while running this function,
        try reducing the ``batch_size`` setting.

    Args:

        source_criteria (bodyId(s), type/instance, or :py:class:`.NeuronCriteria`):
            Criteria to by which to filter source (pre-synaptic) neurons.
            If omitted, all Neurons will be considered as possible sources.

            Note:
                Any ROI criteria specified in this argument does not affect
                which synapses are returned, only which bodies are inspected.

        target_criteria (bodyId(s), type/instance, or :py:class:`.NeuronCriteria`):
            Criteria to by which to filter target (post-synaptic) neurons.
            If omitted, all Neurons will be considered as possible targets.

            Note:
                Any ROI criteria specified in this argument does not affect
                which synapses are returned, only which bodies are inspected.

        synapse_criteria (SynapseCriteria):
            Optional. Allows you to filter synapses by roi or confidence.
            The same criteria is used to filter both ``pre`` and ``post`` sides
            of the connection, except for the ``rois`` -- see note below.
            By default, ``SynapseCriteria(primary_only=True)`` is used,
            with no additional filters.

            If ``primary_only`` is specified in the criteria, then the resulting
            ``roi_pre`` and ``roi_post`` columns will contain a single
            string (or ``None``) in every row.

            Otherwise, the roi columns will contain a list of ROIs for every row.
            (Primary ROIs do not overlap, so every synapse resides in only one
            (or zero) primary ROI.)
            See :py:class:`.SynapseCriteria` for details.

            Note:
                Any ``rois`` specified in your ``synapse_criteria`` will be used to filter
                the target (post-synaptic) side of the synapse connection, but not
                the pre-synaptic side.  So in the rare cases where the pre and post synapses
                reside on different sides of an ROI boundary, the connection is associated
                with the post-synaptic ROI.

                That's consistent with neuprint's conventions for ROI assignment in the
                neuron-to-neuron ``ConnectsTo:`` relationship, and thus ensures that this function
                returns synapse counts that are consistent with :py:func:`.fetch_adjacencies()`.


        min_total_weight:
            If the total weight of the connection between two bodies is not at least
            this strong, don't include the synapses for that connection in the results.

            Note:
                This filters for total connection weight, regardless of the weight
                within any particular ROI.  So, if your ``SynapseCriteria`` limits
                results to a particular ROI, but two bodies connect in multiple ROIs,
                then the number of synapses returned for the two bodies may appear to
                be less than ``min_total_weight``. That's because you filtered out
                the synapses in other ROIs.

        batch_size:
            To avoid timeouts and improve performance, the synapse connections
            will be fetched in batches.  The batching strategy is to process
            each body one at a time, and if it has lots of connection partners,
            split the request across several batches to avoid timeouts that
            could arise from a large request.
            This argument specifies the maximum size of each batch in the inner loop.
            Larger batches are more efficient to fetch, but increase the likelihood
            of timeouts.

        client:
            If not provided, the global default :py:class:`.Client` will be used.

    Returns:

        DataFrame in which each row represents a single synaptic connection
        between an upstream (pre-synaptic) body and downstream (post-synaptic) body.

        Synapse locations are listed in columns ``[x_pre, y_pre, z_pre]`` and
        ``[x_post, y_post, z_post]`` for the upstream and downstream synapses,
        respectively.

        The ``roi_pre`` and ``roi_post`` columns will contain either strings
        or lists-of-strings, depending on the ``primary_only`` synapse criteria as
        described above.

    Example:

        .. code-block:: ipython

            In [1]: from neuprint import fetch_synapse_connections, SynapseCriteria as SC
               ...: fetch_synapse_connections(792368888, None, SC(rois=['PED(R)', 'SMP(R)'], primary_only=True))
            Out[1]:
                bodyId_pre  bodyId_post roi_pre roi_post  x_pre  y_pre  z_pre  x_post  y_post  z_post  confidence_pre  confidence_post
            0    792368888    754547386  PED(R)   PED(R)  14013  27747  19307   13992   27720   19313           0.996         0.401035
            1    792368888    612742248  PED(R)   PED(R)  14049  27681  19417   14044   27662   19408           0.921         0.881487
            2    792368888   5901225361  PED(R)   PED(R)  14049  27681  19417   14055   27653   19420           0.921         0.436177
            3    792368888   5813117385  SMP(R)   SMP(R)  23630  29443  16297   23634   29437   16279           0.984         0.970746
            4    792368888   5813083733  SMP(R)   SMP(R)  23630  29443  16297   23634   29419   16288           0.984         0.933871
            5    792368888   5813058320  SMP(R)   SMP(R)  18662  34144  12692   18655   34155   12697           0.853         0.995000
            6    792368888   5812981989  PED(R)   PED(R)  14331  27921  20099   14351   27928   20085           0.904         0.877373
            7    792368888   5812981381  PED(R)   PED(R)  14331  27921  20099   14301   27919   20109           0.904         0.567321
            8    792368888   5812981381  PED(R)   PED(R)  14013  27747  19307   14020   27747   19285           0.996         0.697836
            9    792368888   5812979314  PED(R)   PED(R)  14331  27921  20099   14329   27942   20109           0.904         0.638362
            10   792368888    424767514  PED(R)   PED(R)  14331  27921  20099   14324   27934   20085           0.904         0.985734
            11   792368888    424767514  PED(R)   PED(R)  14013  27747  19307   14020   27760   19294           0.996         0.942831
            12   792368888    424767514  PED(R)   PED(R)  14049  27681  19417   14040   27663   19420           0.921         0.993586
            13   792368888    331662710  SMP(R)   SMP(R)  23630  29443  16297   23644   29429   16302           0.984         0.996389
            14   792368888   1196854070  PED(R)   PED(R)  14331  27921  20099   14317   27935   20101           0.904         0.968408
            15   792368888   1131831702  SMP(R)   SMP(R)  23630  29443  16297   23651   29434   16316           0.984         0.362952
    """
    assert source_criteria is not None or target_criteria is not None, \
        "Please specify either source or target search criteria (or both)."

    if synapse_criteria is None:
        synapse_criteria = SynapseCriteria(primary_only=True)

    def prepare_nc(nc, matchvar):
        nc.matchvar = matchvar

        # If the user specified rois to filter synapses by, but hasn't specified rois
        # in the NeuronCriteria, add them to the NeuronCriteria to speed up the query.
        if synapse_criteria.rois and not nc.rois:
            nc.rois = {*synapse_criteria.rois}
            nc.roi_req = 'any'

        return nc

    source_criteria = prepare_nc(source_criteria, 'n')
    target_criteria = prepare_nc(target_criteria, 'm')

    # Fetch the list of neuron-neuron pairs in advance so we can break into batches.
    _neuron_df, roi_conn_df = fetch_adjacencies( source_criteria,
                                                 target_criteria,
                                                 synapse_criteria.rois,
                                                 1,
                                                 min_total_weight,
                                                 properties=[] )

    if len(roi_conn_df) == 0:
        # Return empty dataframe, but with the correct dtypes
        dtypes = {
            'bodyId_pre': np.dtype('int64'),
            'bodyId_post': np.dtype('int64'),
            'roi_pre': np.dtype('O'),
            'roi_post': np.dtype('O'),
            'x_pre': np.dtype('int32'),
            'y_pre': np.dtype('int32'),
            'z_pre': np.dtype('int32'),
            'x_post': np.dtype('int32'),
            'y_post': np.dtype('int32'),
            'z_post': np.dtype('int32'),
            'confidence_pre': np.dtype('float32'),
            'confidence_post': np.dtype('float32')
        }
        return pd.DataFrame([], columns=dtypes.keys()).astype(dtypes)

    conn_df = (roi_conn_df.drop_duplicates(['bodyId_pre', 'bodyId_post'])
                          .sort_values(['bodyId_pre', 'bodyId_post']))

    # Pick either 'pre' or 'post' column to process bodies one at a time.
    # Then batch across the connections in an inner loop.
    num_pre = conn_df['bodyId_pre'].nunique()
    num_post = conn_df['bodyId_post'].nunique()
    if num_pre < num_post:
        grouping_col = 'bodyId_pre'
    else:
        grouping_col = 'bodyId_post'

    syn_dfs = []
    with tqdm(total=roi_conn_df['weight'].sum()) as progress:
        for _, group_df in conn_df.groupby(grouping_col):
            batches = iter_batches(group_df, batch_size)
            for batch_df in tqdm(batches, leave=False):
                src_crit = copy.copy(source_criteria)
                tgt_crit = copy.copy(target_criteria)

                if grouping_col == 'bodyId_pre':
                    assert batch_df['bodyId_pre'].nunique() == 1
                    src_crit.bodyId = batch_df['bodyId_pre'].unique()
                    # Filter target criteria further only if connections
                    # are being fetched in multiple batches.
                    # Otherwise, the explicit body list is unnecessary and slows down the query.
                    if len(batches) > 1:
                        tgt_crit.bodyId = batch_df['bodyId_post'].unique()
                else:
                    assert batch_df['bodyId_post'].nunique() == 1
                    tgt_crit.bodyId = batch_df['bodyId_post'].unique()
                    # Filter source criteria further only if connections
                    # are being fetched in multiple batches.
                    # Otherwise, the explicit body list is unnecessary and slows down the query.
                    if len(batches) > 1:
                        src_crit.bodyId = batch_df['bodyId_pre'].unique()

                batch_syn_df = _fetch_synapse_connections( src_crit,
                                                           tgt_crit,
                                                           synapse_criteria,
                                                           min_total_weight,
                                                           client )
                syn_dfs.append(batch_syn_df)
                progress.update(len(batch_syn_df))

    syn_df = pd.concat(syn_dfs, ignore_index=True)
    hashable_cols = [col for col, dtype in syn_df.dtypes.items() if dtype != object]
    assert syn_df.duplicated(hashable_cols).sum() == 0, \
        "Somehow obtained duplicate synapse-synapse connections!"
    return syn_df


def _fetch_synapse_connections(source_criteria, target_criteria, synapse_criteria, min_total_weight, client):
    if synapse_criteria.primary_only:
        return_rois = {*client.primary_rois}
    else:
        return_rois = {*client.all_rois}

    source_syn_crit = copy.copy(synapse_criteria)
    target_syn_crit = copy.copy(synapse_criteria)

    source_syn_crit.matchvar = 'ns'
    target_syn_crit.matchvar = 'ms'

    source_syn_crit.type = 'pre'
    target_syn_crit.type = 'post'

    # We apply ROI filtering to the PSD side only,
    # for consistency with the way neuprint assigns ROIs to
    # connections in the neuron ``ConnectsTo:`` relationship.
    # That way, this function and ``fetch_adjacencies()`` return consistent results.
    source_syn_crit.rois = None

    criteria_globals = [*source_criteria.global_vars().keys(), *target_criteria.global_vars().keys()]
    combined_conditions = NeuronCriteria.combined_conditions(
        [source_criteria, target_criteria],
        ['n', 'e', 'm', 'ns', 'ms', *criteria_globals],
        prefix=8)

    # Fetch results
    cypher = dedent(f"""\
        {NeuronCriteria.combined_global_with((source_criteria, target_criteria), prefix=8)}
        MATCH (n:{source_criteria.label})-[e:ConnectsTo]->(m:{target_criteria.label}),
              (n)-[:Contains]->(nss:SynapseSet)-[:ConnectsTo]->(mss:SynapseSet)<-[:Contains]-(m),
              (nss)-[:Contains]->(ns:Synapse)-[:SynapsesTo]->(ms:Synapse)<-[:Contains]-(mss)

        {combined_conditions}

        // Note: Semantically, the word 'DISTINCT' is unnecessary here,
        // but its presence makes this query run faster.
        WITH DISTINCT n, m, ns, ms, e
        WHERE e.weight >= {min_total_weight}

        {source_syn_crit.condition('n', 'm', 'ns', 'ms', prefix=8)}
        {target_syn_crit.condition('n', 'm', 'ns', 'ms', prefix=8)}

        RETURN n.bodyId as bodyId_pre,
               m.bodyId as bodyId_post,
               ns.location.x as ux,
               ns.location.y as uy,
               ns.location.z as uz,
               ms.location.x as dx,
               ms.location.y as dy,
               ms.location.z as dz,
               ns.confidence as confidence_pre,
               ms.confidence as confidence_post,
               apoc.map.removeKeys(ns, ['location', 'confidence', 'type']) as info_pre,
               apoc.map.removeKeys(ms, ['location', 'confidence', 'type']) as info_post
    """)
    data = client.fetch_custom(cypher, format='json')['data']

    # Assemble DataFrame
    syn_table = []
    for bodyId_pre, bodyId_post, ux, uy, uz, dx, dy, dz, up_conf, dn_conf, info_pre, info_post in data:
        # Exclude non-primary ROIs if necessary
        pre_rois = return_rois & {*info_pre.keys()}
        post_rois = return_rois & {*info_post.keys()}

        # Intern the ROIs to save RAM
        pre_rois = sorted(map(sys.intern, pre_rois))
        post_rois = sorted(map(sys.intern, post_rois))

        pre_rois = pre_rois or [None]
        post_rois = post_rois or [None]

        # Should be (at most) one ROI when primary_only=True,
        # so only show that one (not a list)
        if synapse_criteria.primary_only:
            pre_rois = pre_rois[0]
            post_rois = post_rois[0]

        syn_table.append((bodyId_pre, bodyId_post, pre_rois, post_rois, ux, uy, uz, dx, dy, dz, up_conf, dn_conf))

    syn_df = pd.DataFrame(syn_table, columns=['bodyId_pre', 'bodyId_post',
                                              'roi_pre', 'roi_post',
                                              'x_pre', 'y_pre', 'z_pre', 'x_post', 'y_post', 'z_post',
                                              'confidence_pre', 'confidence_post'])

    # Save RAM with smaller dtypes
    syn_df['x_pre'] = syn_df['x_pre'].astype(np.int32)
    syn_df['y_pre'] = syn_df['y_pre'].astype(np.int32)
    syn_df['z_pre'] = syn_df['z_pre'].astype(np.int32)
    syn_df['x_post'] = syn_df['x_post'].astype(np.int32)
    syn_df['y_post'] = syn_df['y_post'].astype(np.int32)
    syn_df['z_post'] = syn_df['z_post'].astype(np.int32)
    syn_df['confidence_pre'] = syn_df['confidence_pre'].astype(np.float32)
    syn_df['confidence_post'] = syn_df['confidence_post'].astype(np.float32)

    return syn_df
